# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime, timedelta
from typing import Any, Optional
import json
import logging
import os
import sqlite3
import threading
import time


class SQLiteCache:
    """
    A thread-safe SQLite-backed cache for storing prompt responses.

    Uses a persistent connection protected by a threading lock.

    Parameters
    ----------
    db_path : str
        The path to the SQLite database file.
    max_size : int, optional
        The maximum number of items in the cache.
    expiration : timedelta, optional
        The duration after which an item expires.
    """

    def __init__(
        self,
        db_path: str,
        max_size: int = 1000,
        expiration: timedelta = timedelta(days=7),
    ):
        self.db_path = str(db_path)
        self.max_size = max_size
        self.expiration = expiration
        self.lock = threading.Lock()
        # Persistent connection (thread-safe via self.lock)
        self.conn = sqlite3.connect(
            self.db_path, check_same_thread=False
        )
        self._initialize_cache()
        self._set_file_permissions()

    def _set_file_permissions(self):
        """Set restrictive permissions (0600) on the database file."""
        try:
            os.chmod(self.db_path, 0o600)
        except OSError:
            pass  # May fail on Windows or if file is not owned

    def _initialize_cache(self):
        """Create the cache table and indexes if they don't exist."""
        with self.lock:
            cursor = self.conn.cursor()
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS cache (
                    prompt_hash TEXT PRIMARY KEY,
                    response TEXT,
                    timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
                )
                """
            )
            cursor.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_cache_timestamp
                ON cache(timestamp)
                """
            )
            self.conn.commit()
        logging.info(
            "Cache initialized",
            extra={
                "event": "Cache:Initialized",
                "extra_data": {
                    "db_path": self.db_path,
                    "max_size": self.max_size,
                },
            },
        )

    def get(self, prompt_hash: str) -> Optional[str]:
        """
        Retrieve a response from the cache.

        Parameters
        ----------
        prompt_hash : str
            The hash of the prompt.

        Returns
        -------
        Optional[str]
            The cached response, or None if not found/expired.
        """
        start_time = time.time()
        with self.lock:
            cursor = self.conn.cursor()
            cursor.execute(
                """
                SELECT response
                FROM cache
                WHERE prompt_hash = ?
                AND timestamp > ?
                """,
                (prompt_hash, datetime.now() - self.expiration),
            )
            result = cursor.fetchone()
        hit = result is not None
        latency = (time.time() - start_time) * 1000
        logging.info(
            f"Cache {'hit' if hit else 'miss'}",
            extra={
                "event": "Cache:Accessed",
                "extra_data": {
                    "prompt_hash": prompt_hash,
                    "hit": hit,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        if result:
            return json.loads(result[0])
        return None

    def set(self, prompt_hash: str, response: Any) -> None:
        """
        Store a response in the cache.

        Parameters
        ----------
        prompt_hash : str
            The hash of the prompt.
        response : Any
            The response to store.
        """
        start_time = time.time()
        serialized_response = json.dumps(response)
        with self.lock:
            cursor = self.conn.cursor()
            cursor.execute(
                """REPLACE INTO cache (
                    prompt_hash,
                    response,
                    timestamp
                ) VALUES (?, ?, CURRENT_TIMESTAMP)""",
                (prompt_hash, serialized_response),
            )
            # Only evict when over capacity
            cursor.execute("SELECT count(*) FROM cache")
            count = cursor.fetchone()[0]
            if count > self.max_size:
                cursor.execute(
                    """
                    DELETE FROM cache
                    WHERE timestamp <= (
                        SELECT timestamp
                        FROM cache
                        ORDER BY timestamp DESC
                        LIMIT 1 OFFSET ?
                    )
                    """,
                    (self.max_size - 1,),
                )
            self.conn.commit()
        latency = (time.time() - start_time) * 1000
        logging.info(
            "Cache store",
            extra={
                "event": "Cache:Written",
                "extra_data": {
                    "prompt_hash": prompt_hash,
                    "latency_ms": round(latency, 2),
                },
            },
        )

    def close(self):
        """Close the persistent database connection."""
        with self.lock:
            if self.conn:
                self.conn.close()
                self.conn = None



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from ._openai_compat import OpenAICompatProvider


class OpenAIProvider(OpenAICompatProvider):
    """OpenAI API provider.

    Env vars: OPENAI_API_KEY, OPENAI_DEFAULT_MODEL
    """

    _provider_name = "openai"

    def __init__(self):
        from akande.config import (
            OPENAI_API_KEY,
            OPENAI_DEFAULT_MODEL,
        )

        self._api_key = OPENAI_API_KEY or ""
        self._base_url = ""
        self._default_model = (
            OPENAI_DEFAULT_MODEL or "gpt-3.5-turbo"
        )
        self._init_client()



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

from ._openai_compat import OpenAICompatProvider


class OllamaProvider(OpenAICompatProvider):
    """Ollama local inference provider (OpenAI-compatible API).

    Env vars: OLLAMA_HOST (default: http://localhost:11434)
    """

    _provider_name = "ollama"

    def __init__(self):
        host = os.getenv(
            "OLLAMA_HOST", "http://localhost:11434"
        )
        self._api_key = "ollama"
        self._base_url = f"{host}/v1"
        self._default_model = "llama3"
        self._init_client()



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import logging
import os
import time
from typing import Any, Dict, Optional

from .base import LLMProvider
from .response import ProviderResponse


class HuggingFaceProvider(LLMProvider):
    """Hugging Face Inference API provider.

    Env vars: HUGGINGFACE_API_KEY
    """

    @property
    def provider_name(self) -> str:
        return "huggingface"

    def __init__(self):
        try:
            from huggingface_hub import InferenceClient
        except ImportError:
            raise ImportError(
                "The 'huggingface_hub' package is required for "
                "the Hugging Face provider. Install it with: "
                "pip install akande[huggingface]"
            )
        api_key = os.getenv("HUGGINGFACE_API_KEY", "")
        if not api_key:
            raise ValueError(
                "HUGGINGFACE_API_KEY environment variable "
                "is required for the Hugging Face provider."
            )
        self.client = InferenceClient(token=api_key)
        self._default_model = (
            "mistralai/Mistral-7B-Instruct-v0.2"
        )

    def _call(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> ProviderResponse:
        if not params:
            params = {}
        model = model or self._default_model
        response = self.client.chat_completion(
            model=model,
            messages=[
                {
                    "role": "system",
                    "content": system_prompt,
                },
                {
                    "role": "user",
                    "content": user_prompt,
                },
            ],
            max_tokens=1024,
            **params,
        )
        text = response.choices[0].message.content
        return ProviderResponse(text)

    async def generate_response(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        logging.info(
            "LLM request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": "huggingface",
                    "model": model or self._default_model,
                },
            },
        )
        start = time.time()
        try:
            loop = asyncio.get_running_loop()
            response = await loop.run_in_executor(
                None,
                lambda: self._call(
                    user_prompt,
                    system_prompt,
                    model,
                    params,
                ),
            )
        except Exception:
            latency = (time.time() - start) * 1000
            logging.error(
                "LLM request failed",
                exc_info=True,
                extra={
                    "event": "LLM:RequestFailed",
                    "extra_data": {
                        "provider": "huggingface",
                        "model": (
                            model or self._default_model
                        ),
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            raise
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": "huggingface",
                    "model": model or self._default_model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response

    def generate_response_sync(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        logging.info(
            "LLM sync request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": "huggingface",
                    "model": model or self._default_model,
                },
            },
        )
        start = time.time()
        try:
            response = self._call(
                user_prompt, system_prompt, model, params
            )
        except Exception:
            latency = (time.time() - start) * 1000
            logging.error(
                "LLM sync request failed",
                exc_info=True,
                extra={
                    "event": "LLM:RequestFailed",
                    "extra_data": {
                        "provider": "huggingface",
                        "model": (
                            model or self._default_model
                        ),
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            raise
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM sync response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": "huggingface",
                    "model": model or self._default_model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class ProviderMessage:
    """Normalised message matching OpenAI's response format."""

    __slots__ = ("content",)

    def __init__(self, content: str):
        self.content = content


class ProviderChoice:
    """Normalised choice matching OpenAI's response format."""

    __slots__ = ("message",)

    def __init__(self, message: ProviderMessage):
        self.message = message


class ProviderResponse:
    """Normalised response wrapper for non-OpenAI providers.

    Provides a `.choices[0].message.content` interface matching
    the OpenAI SDK response format so that the rest of the
    application can consume any provider's output uniformly.
    """

    __slots__ = ("choices",)

    def __init__(self, content: str):
        self.choices = [
            ProviderChoice(ProviderMessage(content))
        ]



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

from ._openai_compat import OpenAICompatProvider


class AzureOpenAIProvider(OpenAICompatProvider):
    """Azure OpenAI Service provider.

    Env vars: AZURE_OPENAI_API_KEY, AZURE_OPENAI_ENDPOINT,
              AZURE_OPENAI_API_VERSION
    """

    _provider_name = "azure_openai"

    def __init__(self):
        import openai
        from akande.config import API_CALL_TIMEOUT

        api_key = os.getenv("AZURE_OPENAI_API_KEY", "")
        endpoint = os.getenv("AZURE_OPENAI_ENDPOINT", "")
        api_version = os.getenv(
            "AZURE_OPENAI_API_VERSION", "2024-02-01"
        )

        if not api_key:
            raise ValueError(
                "AZURE_OPENAI_API_KEY environment variable "
                "is required for Azure OpenAI provider."
            )
        if not endpoint:
            raise ValueError(
                "AZURE_OPENAI_ENDPOINT environment variable "
                "is required for Azure OpenAI provider."
            )

        self._api_key = api_key
        self._base_url = endpoint
        self._default_model = "gpt-35-turbo"

        self.client = openai.AzureOpenAI(
            api_key=api_key,
            azure_endpoint=endpoint,
            api_version=api_version,
            timeout=API_CALL_TIMEOUT,
        )



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import logging

from .akande import Akande
from .config import LLM_PROVIDER, OPENAI_API_KEY
from .logger import basic_config
from .providers import get_provider
from .services import OpenAIImpl
from .utils import (
    get_output_directory,
    get_output_filename,
    validate_api_key,
)


async def main():
    """
    Main function to initialize and run the Akande voice assistant.
    """
    provider_name = LLM_PROVIDER or "openai"

    # For OpenAI provider, validate the API key format
    if provider_name == "openai" and not validate_api_key(
        OPENAI_API_KEY
    ):
        logging.error(
            "Invalid or missing OPENAI_API_KEY",
            extra={"event": "Config:ValidationFailed"},
        )
        return

    try:
        provider = get_provider(provider_name)
    except (ValueError, ImportError) as e:
        logging.error(
            "Provider initialization failed",
            extra={
                "event": "Provider:InitFailed",
                "extra_data": {
                    "provider": provider_name,
                    "error": type(e).__name__,
                },
            },
        )
        return

    # Use the resolved provider. For backward compatibility,
    # wrap it in a lightweight OpenAIImpl when OpenAI is
    # selected; otherwise use the provider directly.
    if provider_name == "openai":
        openai_service = OpenAIImpl()
    else:
        openai_service = provider
    akande = Akande(openai_service=openai_service)
    try:
        await akande.run_interaction()
    except KeyboardInterrupt:
        logging.info(
            "Keyboard interrupt detected, exiting",
            extra={"event": "Session:Ended"},
        )
        await akande.stop_server()


def run():
    """Synchronous entry point for console_scripts."""
    directory_path = get_output_directory()
    filename = get_output_filename(".log")
    file_path = directory_path / filename

    log_level = logging.INFO
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    basic_config(
        filename=str(file_path),
        level=log_level,
        log_format=log_format,
    )

    asyncio.run(main())


if __name__ == "__main__":
    run()



import json
import time
from unittest.mock import MagicMock, patch

import pytest

from akande.server.server import (
    AkandeServer,
    ALLOWED_STATIC_FILES,
    MAX_QUESTION_LENGTH,
    MAX_AUDIO_SIZE,
    RATE_LIMIT_WINDOW,
    RATE_LIMIT_MAX_REQUESTS,
    RateLimiter,
    _detect_audio_format,
    _hash_ip,
)


class TestAudioFormatDetection:
    def test_detect_webm(self):
        data = b"\x1a\x45\xdf\xa3" + b"\x00" * 100
        assert _detect_audio_format(data) == "webm"

    def test_detect_mp3_id3(self):
        data = b"ID3" + b"\x00" * 100
        assert _detect_audio_format(data) == "mp3"

    def test_detect_ogg(self):
        data = b"OggS" + b"\x00" * 100
        assert _detect_audio_format(data) == "ogg"

    def test_detect_flac(self):
        data = b"fLaC" + b"\x00" * 100
        assert _detect_audio_format(data) == "flac"

    def test_detect_mp4_at_offset(self):
        data = b"\x00\x00\x00\x20ftyp" + b"\x00" * 100
        assert _detect_audio_format(data) == "mp4"

    def test_detect_unknown(self):
        data = b"\x00\x00\x00\x00\x00" * 20
        assert _detect_audio_format(data) == ""


class TestRateLimiter:
    def test_allows_within_limit(self):
        limiter = RateLimiter(window=60, max_requests=5)
        for _ in range(5):
            assert limiter.is_allowed("127.0.0.1") is True

    def test_blocks_over_limit(self):
        limiter = RateLimiter(window=60, max_requests=3)
        for _ in range(3):
            limiter.is_allowed("127.0.0.1")
        assert limiter.is_allowed("127.0.0.1") is False

    def test_different_ips_independent(self):
        limiter = RateLimiter(window=60, max_requests=1)
        assert limiter.is_allowed("1.1.1.1") is True
        assert limiter.is_allowed("2.2.2.2") is True
        assert limiter.is_allowed("1.1.1.1") is False

    def test_cleanup_removes_stale_entries(self):
        limiter = RateLimiter(window=1, max_requests=10)
        limiter.is_allowed("1.1.1.1")
        limiter.is_allowed("2.2.2.2")
        time.sleep(1.1)
        limiter.cleanup()
        assert "1.1.1.1" not in limiter._requests
        assert "2.2.2.2" not in limiter._requests

    def test_thread_safe(self):
        """Rate limiter should work under concurrent access."""
        import threading

        limiter = RateLimiter(window=60, max_requests=100)
        results = []

        def make_requests():
            for _ in range(50):
                results.append(limiter.is_allowed("127.0.0.1"))

        threads = [
            threading.Thread(target=make_requests)
            for _ in range(4)
        ]
        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # Should have allowed exactly 100
        assert sum(results) == 100
        assert len(results) == 200


class TestHashIp:
    def test_hashes_ip(self):
        result = _hash_ip("127.0.0.1")
        assert isinstance(result, str)
        assert len(result) == 12

    def test_deterministic(self):
        assert _hash_ip("1.2.3.4") == _hash_ip("1.2.3.4")

    def test_different_ips_different_hashes(self):
        assert _hash_ip("1.2.3.4") != _hash_ip("5.6.7.8")


class TestAkandeServerInit:
    @patch(
        "akande.server.server.LLM_PROVIDER", "openai"
    )
    @patch(
        "akande.server.server.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.server.server.OpenAIImpl")
    def test_init(self, mock_impl):
        server = AkandeServer()
        assert server.openai_service is not None
        assert server.public_dir is not None
        assert server.cache is not None

    @patch(
        "akande.server.server.LLM_PROVIDER", "openai"
    )
    @patch("akande.server.server.OPENAI_API_KEY", None)
    def test_init_fails_without_api_key(self):
        with pytest.raises(
            RuntimeError, match="OPENAI_API_KEY"
        ):
            AkandeServer()


class TestStaticFileServing:
    @patch(
        "akande.server.server.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.server.server.OpenAIImpl")
    def test_forbidden_path(self, mock_impl):
        import cherrypy

        server = AkandeServer()
        with pytest.raises(cherrypy.HTTPError) as exc_info:
            server.static("../../etc/passwd")
        assert exc_info.value.status == 403

    @patch(
        "akande.server.server.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.server.server.OpenAIImpl")
    def test_non_allowed_file(self, mock_impl):
        import cherrypy

        server = AkandeServer()
        with pytest.raises(cherrypy.HTTPError) as exc_info:
            server.static("secret.txt")
        assert exc_info.value.status == 403


class TestHealthEndpoint:
    @patch(
        "akande.server.server.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.server.server.OpenAIImpl")
    @patch("akande.server.server.cherrypy")
    def test_health_returns_ok(self, mock_cp, mock_impl):
        mock_cp.response = MagicMock()
        mock_cp.response.headers = {}
        server = AkandeServer()
        result = server.health()
        data = json.loads(result)
        assert data["status"] == "ok"
        assert data["service"] == "akande"


class TestProcessAudio:
    def test_process_audio_unknown_value(self, tmp_path):
        import wave

        wav_path = tmp_path / "test.wav"
        with wave.open(str(wav_path), "w") as wf:
            wf.setnchannels(1)
            wf.setsampwidth(2)
            wf.setframerate(16000)
            wf.writeframes(b"\x00" * 32000)

        with patch(
            "akande.server.server.sr.Recognizer"
        ) as mock_recognizer_cls:
            mock_recognizer = MagicMock()
            mock_recognizer_cls.return_value = mock_recognizer

            from speech_recognition import UnknownValueError

            # Patch the module-level _recognizer
            with patch(
                "akande.server.server._recognizer"
            ) as mock_rec:
                mock_rec.recognize_google.side_effect = (
                    UnknownValueError()
                )

                result = AkandeServer.process_audio(
                    str(wav_path)
                )
                assert result["success"] is False
                assert "could not be understood" in result["error"]


class TestConstants:
    def test_allowed_static_files(self):
        assert "sine-wave-generator.js" in ALLOWED_STATIC_FILES

    def test_max_question_length(self):
        assert MAX_QUESTION_LENGTH == 5000

    def test_max_audio_size(self):
        assert MAX_AUDIO_SIZE == 10 * 1024 * 1024

    def test_rate_limit_window(self):
        assert RATE_LIMIT_WINDOW == 60

    def test_rate_limit_max_requests(self):
        assert RATE_LIMIT_MAX_REQUESTS == 20



import os
from unittest import mock


def test_config_loads_env_vars():
    with mock.patch.dict(
        os.environ,
        {
            "OPENAI_API_KEY": "sk-test-key-1234567890abcdef",
            "OPENAI_DEFAULT_MODEL": "gpt-4",
        },
    ):
        # Re-import to pick up mocked env
        import importlib
        from akande import config

        importlib.reload(config)
        assert config.OPENAI_API_KEY == "sk-test-key-1234567890abcdef"
        assert config.OPENAI_DEFAULT_MODEL == "gpt-4"


def test_config_defaults():
    with mock.patch.dict(os.environ, {}, clear=True):
        import importlib
        from akande import config

        importlib.reload(config)
        assert config.OPENAI_API_KEY is None
        assert config.OPENAI_DEFAULT_MODEL is None


def test_api_call_timeout():
    from akande.config import API_CALL_TIMEOUT

    assert API_CALL_TIMEOUT == 90



import csv
from pathlib import Path

from akande.utils import (
    generate_csv,
    generate_pdf,
    validate_api_key,
    get_output_directory,
    get_output_filename,
)


class TestValidateApiKey:
    def test_valid_key(self):
        assert validate_api_key("sk-abcdefghijklmnopqrstuvwxyz") is True

    def test_valid_proj_key(self):
        assert (
            validate_api_key("sk-proj-abcdefghijklmnopqrstuvwxyz")
            is True
        )

    def test_valid_org_key(self):
        assert (
            validate_api_key("sk-org-abcdefghijklmnopqrstuvwxyz")
            is True
        )

    def test_none_key(self):
        assert validate_api_key(None) is False

    def test_empty_key(self):
        assert validate_api_key("") is False

    def test_short_key(self):
        assert validate_api_key("sk-short") is False

    def test_wrong_prefix(self):
        assert (
            validate_api_key("wrong-prefix-abcdefghijklmnop")
            is False
        )

    def test_almost_valid_prefix(self):
        assert (
            validate_api_key("ska-abcdefghijklmnopqrstuvwxyz")
            is False
        )


class TestGetOutputDirectory:
    def test_creates_directory(self, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = get_output_directory()
        assert result.exists()
        assert result.is_dir()

    def test_returns_path(self, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        result = get_output_directory()
        assert isinstance(result, Path)


class TestGetOutputFilename:
    def test_includes_extension(self):
        result = get_output_filename(".pdf")
        assert result.endswith(".pdf")

    def test_includes_seconds(self):
        result = get_output_filename(".csv")
        # Format is YYYY-MM-DD-HH-MM-SS-Akande.csv
        parts = result.replace("-Akande.csv", "").split("-")
        # Should have 6 parts: year, month, day, hour, min, sec
        assert len(parts) == 6

    def test_includes_akande(self):
        result = get_output_filename(".wav")
        assert "Akande" in result


class TestGenerateCsv:
    def test_generates_csv_file(self, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        generate_csv("test question", "test response")

        csv_files = list(tmp_path.rglob("*.csv"))
        assert len(csv_files) == 1

        with open(csv_files[0], newline="", encoding="utf-8") as f:
            reader = csv.reader(f)
            rows = list(reader)
            assert rows[0] == ["Question", "Response"]
            assert rows[1] == ["test question", "test response"]

    def test_handles_error_gracefully(self, tmp_path, monkeypatch):
        """generate_csv should log errors, not raise."""
        monkeypatch.chdir(tmp_path)
        # Make directory read-only to force write error
        import os

        read_only_dir = tmp_path / "readonly"
        read_only_dir.mkdir()
        os.chmod(str(read_only_dir), 0o444)
        # This won't crash because generate_csv creates
        # its own date dir in cwd, not in readonly dir
        generate_csv("q", "r")  # Should not raise


class TestGeneratePdf:
    def test_generates_pdf_file(self, tmp_path, monkeypatch):
        monkeypatch.chdir(tmp_path)
        generate_pdf("test question", "Overview\nSome text")

        pdf_files = list(tmp_path.rglob("*.pdf"))
        assert len(pdf_files) == 1

    def test_generates_pdf_without_logo(
        self, tmp_path, monkeypatch
    ):
        monkeypatch.chdir(tmp_path)
        generate_pdf("question", "response text")

        pdf_files = list(tmp_path.rglob("*.pdf"))
        assert len(pdf_files) == 1

    def test_generates_pdf_with_sections(
        self, tmp_path, monkeypatch
    ):
        monkeypatch.chdir(tmp_path)
        response = (
            "Overview\nIntro text\n"
            "Solution\n- step one\n- step two\n"
            "Conclusion\nWrap up\n"
            "Recommendations\nDo this"
        )
        generate_pdf("question", response)

        pdf_files = list(tmp_path.rglob("*.pdf"))
        assert len(pdf_files) == 1

    def test_escapes_markup_in_question(
        self, tmp_path, monkeypatch
    ):
        """User input with XML/HTML should be escaped."""
        monkeypatch.chdir(tmp_path)
        # This should not crash even with markup chars
        generate_pdf(
            "<script>alert('xss')</script>",
            "Safe response & <ok>",
        )
        pdf_files = list(tmp_path.rglob("*.pdf"))
        assert len(pdf_files) == 1



import os
from datetime import timedelta

import pytest

from akande.cache import SQLiteCache


@pytest.fixture
def cache(tmp_path):
    db_path = tmp_path / "test_cache.db"
    return SQLiteCache(db_path, max_size=5, expiration=timedelta(days=1))


def test_set_and_get(cache):
    cache.set("hash1", "response1")
    result = cache.get("hash1")
    assert result == "response1"


def test_get_missing_key(cache):
    result = cache.get("nonexistent")
    assert result is None


def test_cache_overwrite(cache):
    cache.set("hash1", "response1")
    cache.set("hash1", "response2")
    result = cache.get("hash1")
    assert result == "response2"


def test_cache_eviction(cache):
    # Cache max_size is 5; insert 6 items
    for i in range(6):
        cache.set(f"hash{i}", f"response{i}")

    # The oldest entry should have been evicted
    # (exact eviction depends on timestamp granularity)
    stored_count = 0
    for i in range(6):
        if cache.get(f"hash{i}") is not None:
            stored_count += 1
    assert stored_count <= 5


def test_cache_expiration(tmp_path):
    db_path = tmp_path / "test_expire.db"
    cache = SQLiteCache(
        db_path, max_size=10, expiration=timedelta(seconds=-1)
    )
    cache.set("hash1", "response1")
    # With negative expiration, everything is expired
    result = cache.get("hash1")
    assert result is None


def test_cache_stores_dict(cache):
    data = {"key": "value", "number": 42}
    cache.set("hash_dict", data)
    result = cache.get("hash_dict")
    assert result == data


def test_cache_stores_list(cache):
    data = [1, 2, 3, "four"]
    cache.set("hash_list", data)
    result = cache.get("hash_list")
    assert result == data


def test_cache_thread_safety(cache):
    import threading

    errors = []

    def writer(key):
        try:
            cache.set(key, f"value_{key}")
        except Exception as e:
            errors.append(e)

    threads = [
        threading.Thread(target=writer, args=(f"k{i}",))
        for i in range(10)
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert len(errors) == 0


def test_cache_close(tmp_path):
    db_path = tmp_path / "test_close.db"
    cache = SQLiteCache(db_path)
    cache.set("k", "v")
    cache.close()
    assert cache.conn is None


def test_cache_file_permissions(tmp_path):
    db_path = tmp_path / "test_perms.db"
    cache = SQLiteCache(db_path)
    stat = os.stat(str(db_path))
    # Check owner-only read/write (0600)
    assert oct(stat.st_mode)[-3:] == "600"
    cache.close()


def test_cache_has_timestamp_index(cache):
    """Verify the timestamp index exists."""
    with cache.lock:
        cursor = cache.conn.cursor()
        cursor.execute(
            "SELECT name FROM sqlite_master "
            "WHERE type='index' AND name='idx_cache_timestamp'"
        )
        result = cursor.fetchone()
    assert result is not None



import asyncio
from unittest.mock import MagicMock, AsyncMock, patch

from akande.akande import (
    Akande,
    Colors,
    CLEAR_SCREEN,
    MAX_THREAD_WORKERS,
    TTS_RATE,
    CACHE_DB_NAME,
)


class TestColors:
    def test_reset_defined(self):
        assert Colors.RESET == "\033[0m"

    def test_header_defined(self):
        assert Colors.HEADER == "\033[95m"


class TestConstants:
    def test_max_thread_workers(self):
        assert MAX_THREAD_WORKERS == 4

    def test_tts_rate(self):
        assert TTS_RATE == 161

    def test_cache_db_name(self):
        assert CACHE_DB_NAME == "akande_cache.db"

    def test_clear_screen_is_ansi(self):
        assert CLEAR_SCREEN == "\033[2J\033[H"


class TestAkandeInit:
    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_init(self, mock_recognizer, mock_cache, mock_tts):
        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        assert akande.openai_service == mock_service
        assert akande.server is None
        assert akande.server_running is False


class TestHashPrompt:
    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_hash_deterministic(
        self, mock_recognizer, mock_cache, mock_tts
    ):
        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        hash1 = akande.hash_prompt("test")
        hash2 = akande.hash_prompt("test")
        assert hash1 == hash2

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_hash_different_inputs(
        self, mock_recognizer, mock_cache, mock_tts
    ):
        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        hash1 = akande.hash_prompt("test1")
        hash2 = akande.hash_prompt("test2")
        assert hash1 != hash2


class TestGenerateResponse:
    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_cache_hit(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = "cached response"
        mock_cache_cls.return_value = mock_cache

        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        result = asyncio.run(akande.generate_response("test"))
        assert result == "cached response"
        mock_service.generate_response.assert_not_called()

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_cache_miss(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = None
        mock_cache_cls.return_value = mock_cache

        mock_response = MagicMock()
        mock_response.choices = [
            MagicMock(
                message=MagicMock(content="  new response  ")
            )
        ]

        mock_service = MagicMock()
        mock_service.generate_response = AsyncMock(
            return_value=mock_response
        )

        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        result = asyncio.run(akande.generate_response("test"))
        assert result == "new response"
        mock_cache.set.assert_called_once()

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_api_error_returns_empty(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = None
        mock_cache_cls.return_value = mock_cache

        mock_service = MagicMock()
        mock_service.generate_response = AsyncMock(
            side_effect=Exception("API down")
        )

        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        result = asyncio.run(akande.generate_response("test"))
        assert result == ""

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_unexpected_response_returns_empty(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = None
        mock_cache_cls.return_value = mock_cache

        # Return a dict (no .choices attribute)
        mock_service = MagicMock()
        mock_service.generate_response = AsyncMock(
            return_value={"error": "bad"}
        )

        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        result = asyncio.run(akande.generate_response("test"))
        assert result == ""

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_correlation_id_passed(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = "cached"
        mock_cache_cls.return_value = mock_cache

        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        result = asyncio.run(
            akande.generate_response(
                "test", correlation_id="test-123"
            )
        )
        assert result == "cached"


class TestServerLifecycle:
    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_run_server_sets_running(
        self, mock_recognizer, mock_cache, mock_tts
    ):
        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)

        with patch("akande.akande.cherrypy"):
            asyncio.run(akande.run_server())
            assert akande.server_running is True
            assert akande.server_thread is not None

    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_run_server_noop_when_running(
        self, mock_recognizer, mock_cache, mock_tts
    ):
        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        akande.server_running = True

        asyncio.run(akande.run_server())
        assert akande.server_thread is None


class TestPIIProtection:
    @patch("akande.akande.pyttsx4")
    @patch("akande.akande.SQLiteCache")
    @patch("akande.akande.sr.Recognizer")
    def test_cache_hit_logs_hash_not_prompt(
        self, mock_recognizer, mock_cache_cls, mock_tts
    ):
        mock_cache = MagicMock()
        mock_cache.get.return_value = "cached"
        mock_cache_cls.return_value = mock_cache

        mock_service = MagicMock()
        akande = Akande(openai_service=mock_service)
        akande.cache = mock_cache

        with patch("akande.akande.logging") as mock_logging:
            asyncio.run(
                akande.generate_response("my secret prompt")
            )
            log_calls = str(mock_logging.info.call_args_list)
            assert "my secret prompt" not in log_calls



import json
import logging

from akande.logger import basic_config, JSONFormatter


def test_basic_config_creates_file_handler(tmp_path):
    log_file = tmp_path / "test.log"
    root = logging.getLogger()
    for handler in root.handlers[:]:
        root.removeHandler(handler)

    basic_config(
        filename=str(log_file),
        level=logging.DEBUG,
        log_format="%(message)s",
    )

    logging.debug("test message")
    assert log_file.exists()

    # Clean up handlers
    for handler in root.handlers[:]:
        root.removeHandler(handler)


def test_basic_config_adds_console_handler(tmp_path):
    log_file = tmp_path / "test2.log"
    root = logging.getLogger()
    for handler in root.handlers[:]:
        root.removeHandler(handler)

    basic_config(
        filename=str(log_file),
        level=logging.DEBUG,
        log_format="%(message)s",
    )

    handler_types = [type(h).__name__ for h in root.handlers]
    assert "FileHandler" in handler_types
    assert "StreamHandler" in handler_types

    # Clean up handlers
    for handler in root.handlers[:]:
        root.removeHandler(handler)


def test_basic_config_replaces_existing_handlers(tmp_path):
    log_file = tmp_path / "test3.log"
    root = logging.getLogger()
    for handler in root.handlers[:]:
        root.removeHandler(handler)

    # Call twice — should not duplicate handlers
    basic_config(
        filename=str(log_file),
        level=logging.DEBUG,
        log_format="%(message)s",
    )
    basic_config(
        filename=str(log_file),
        level=logging.DEBUG,
        log_format="%(message)s",
    )

    # Should have exactly 2 handlers (file + console)
    assert len(root.handlers) == 2

    # Clean up handlers
    for handler in root.handlers[:]:
        root.removeHandler(handler)


def test_json_formatter_output():
    formatter = JSONFormatter(service="test-service")
    record = logging.LogRecord(
        name="test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Test message",
        args=(),
        exc_info=None,
    )
    output = formatter.format(record)
    data = json.loads(output)
    assert data["level"] == "INFO"
    assert data["service"] == "test-service"
    assert data["message"] == "Test message"
    assert "timestamp" in data


def test_json_formatter_includes_correlation_id():
    formatter = JSONFormatter()
    record = logging.LogRecord(
        name="test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Test",
        args=(),
        exc_info=None,
    )
    record.correlation_id = "abc-123"
    output = formatter.format(record)
    data = json.loads(output)
    assert data["correlation_id"] == "abc-123"


def test_json_formatter_includes_event():
    formatter = JSONFormatter()
    record = logging.LogRecord(
        name="test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Test",
        args=(),
        exc_info=None,
    )
    record.event = "Cache:Accessed"
    output = formatter.format(record)
    data = json.loads(output)
    assert data["event"] == "Cache:Accessed"


def test_json_formatter_includes_extra_data():
    formatter = JSONFormatter()
    record = logging.LogRecord(
        name="test",
        level=logging.INFO,
        pathname="test.py",
        lineno=1,
        msg="Test",
        args=(),
        exc_info=None,
    )
    record.extra_data = {"key": "value"}
    output = formatter.format(record)
    data = json.loads(output)
    assert data["data"]["key"] == "value"


def test_json_formatter_handles_exception():
    formatter = JSONFormatter()
    try:
        raise ValueError("test error")
    except ValueError:
        import sys

        record = logging.LogRecord(
            name="test",
            level=logging.ERROR,
            pathname="test.py",
            lineno=1,
            msg="Error occurred",
            args=(),
            exc_info=sys.exc_info(),
        )
    output = formatter.format(record)
    data = json.loads(output)
    assert data["exception"]["type"] == "ValueError"
    assert data["exception"]["message"] == "test error"


def test_file_handler_uses_json_format(tmp_path):
    log_file = tmp_path / "json_test.log"
    root = logging.getLogger()
    for handler in root.handlers[:]:
        root.removeHandler(handler)

    basic_config(
        filename=str(log_file),
        level=logging.DEBUG,
        log_format="%(message)s",
    )

    logging.info("structured test")

    # Clean up handlers to flush
    for handler in root.handlers[:]:
        handler.flush()
        root.removeHandler(handler)

    content = log_file.read_text().strip()
    # File handler should write JSON
    data = json.loads(content)
    assert data["message"] == "structured test"
    assert data["level"] == "INFO"



"""Tests for akande.providers package.

Tests all 10 provider adapters, the registry, the ABC, the
response normalisation layer, and the OpenAI-compatible base.
"""

import asyncio
from unittest.mock import MagicMock, patch

import pytest

from akande.providers.base import LLMProvider
from akande.providers.response import (
    ProviderResponse,
    ProviderChoice,
    ProviderMessage,
)
from akande.providers.registry import (
    ProviderRegistry,
    get_provider,
    DEFAULT_PROVIDER,
    _registry,
)


@pytest.fixture(autouse=True)
def _clear_registry_cache():
    """Clear cached provider instances between tests."""
    _registry._instances.clear()
    yield
    _registry._instances.clear()


# ────────────────────────────────────────────────────────────
# Response normalisation
# ────────────────────────────────────────────────────────────


class TestProviderResponse:
    """Test the ProviderResponse normalisation wrapper."""

    def test_content_accessible(self):
        r = ProviderResponse("hello world")
        assert r.choices[0].message.content == "hello world"

    def test_empty_content(self):
        r = ProviderResponse("")
        assert r.choices[0].message.content == ""

    def test_message_class(self):
        m = ProviderMessage("text")
        assert m.content == "text"

    def test_choice_class(self):
        c = ProviderChoice(ProviderMessage("x"))
        assert c.message.content == "x"


# ────────────────────────────────────────────────────────────
# LLMProvider ABC
# ────────────────────────────────────────────────────────────


class TestLLMProviderABC:
    """Test that LLMProvider cannot be instantiated directly."""

    def test_cannot_instantiate(self):
        with pytest.raises(TypeError):
            LLMProvider()

    def test_concrete_subclass_works(self):
        class Dummy(LLMProvider):
            @property
            def provider_name(self):
                return "dummy"

            async def generate_response(
                self, user_prompt, system_prompt, model,
                params=None,
            ):
                return "ok"

            def generate_response_sync(
                self, user_prompt, system_prompt, model,
                params=None,
            ):
                return "ok"

        d = Dummy()
        assert d.provider_name == "dummy"
        assert d.generate_response_sync("a", "b", "c") == "ok"


# ────────────────────────────────────────────────────────────
# ProviderRegistry
# ────────────────────────────────────────────────────────────


class TestProviderRegistry:

    def test_register_and_create(self):
        reg = ProviderRegistry()

        class FakeProvider(LLMProvider):
            @property
            def provider_name(self):
                return "fake"

            async def generate_response(
                self, up, sp, m, params=None
            ):
                return None

            def generate_response_sync(
                self, up, sp, m, params=None
            ):
                return None

        reg.register("fake", FakeProvider)
        provider = reg.create("fake")
        assert isinstance(provider, FakeProvider)

    def test_create_unknown_raises(self):
        reg = ProviderRegistry()
        with pytest.raises(
            ValueError, match="Unknown LLM provider"
        ):
            reg.create("nonexistent")

    def test_available_lists_registered(self):
        reg = ProviderRegistry()

        class FP(LLMProvider):
            @property
            def provider_name(self):
                return "fp"

            async def generate_response(
                self, up, sp, m, params=None
            ):
                return None

            def generate_response_sync(
                self, up, sp, m, params=None
            ):
                return None

        reg.register("alpha", FP)
        reg.register("beta", FP)
        assert "alpha" in reg.available
        assert "beta" in reg.available

    def test_create_caches_instance(self):
        reg = ProviderRegistry()

        class FP(LLMProvider):
            @property
            def provider_name(self):
                return "fp"

            async def generate_response(
                self, up, sp, m, params=None
            ):
                return None

            def generate_response_sync(
                self, up, sp, m, params=None
            ):
                return None

        reg.register("cached", FP)
        p1 = reg.create("cached")
        p2 = reg.create("cached")
        assert p1 is p2

    def test_lazy_registration(self):
        reg = ProviderRegistry()
        reg.register_lazy(
            "openai",
            ".openai_provider",
            "OpenAIProvider",
        )
        assert "openai" in reg.available

    def test_global_registry_has_10_providers(self):
        assert len(_registry.available) == 10

    def test_global_registry_provider_names(self):
        expected = {
            "openai", "anthropic", "google", "ollama",
            "azure_openai", "mistral", "cohere",
            "huggingface", "groq", "lmstudio",
        }
        assert set(_registry.available) == expected


# ────────────────────────────────────────────────────────────
# get_provider
# ────────────────────────────────────────────────────────────


class TestGetProvider:

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_get_provider_openai(self, mock_openai_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        provider = get_provider("openai")
        assert isinstance(provider, OpenAIProvider)
        assert provider.provider_name == "openai"

    def test_get_provider_unknown_raises(self):
        with pytest.raises(
            ValueError, match="Unknown LLM provider"
        ):
            get_provider("nonexistent_provider")

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.config.LLM_PROVIDER", "openai")
    def test_get_provider_default_from_config(
        self, mock_openai_cls
    ):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        provider = get_provider()
        assert isinstance(provider, OpenAIProvider)

    @patch("akande.config.LLM_PROVIDER", "")
    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_get_provider_empty_falls_back_to_default(
        self, mock_openai_cls
    ):
        provider = get_provider("")
        assert provider.provider_name == "openai"


class TestDefaultProviderConstant:

    def test_default_is_openai(self):
        assert DEFAULT_PROVIDER == "openai"


# ────────────────────────────────────────────────────────────
# OpenAI Provider
# ────────────────────────────────────────────────────────────


class TestOpenAIProvider:

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        assert OpenAIProvider().provider_name == "openai"

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_generate_response_sync(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = OpenAIProvider()
        result = p.generate_response_sync(
            "hello", "sys", "gpt-4"
        )
        mock_client.chat.completions.create.assert_called_once()
        assert result is not None

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_generate_response_async(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = OpenAIProvider()
        result = asyncio.run(
            p.generate_response("hello", "sys", "gpt-4")
        )
        assert result is not None

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_passes_system_and_user_messages(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = OpenAIProvider()
        p.generate_response_sync("user", "system", "gpt-4")
        args = mock_client.chat.completions.create.call_args
        msgs = args.kwargs.get("messages", [])
        assert msgs[0] == {
            "role": "system", "content": "system"
        }
        assert msgs[1] == {
            "role": "user", "content": "user"
        }

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_passes_extra_params(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = OpenAIProvider()
        p.generate_response_sync(
            "hi", "sys", "gpt-4", params={"temperature": 0.5}
        )
        args = mock_client.chat.completions.create.call_args
        assert args.kwargs.get("temperature") == 0.5

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_DEFAULT_MODEL",
        "gpt-4o",
    )
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_respects_default_model_env(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        p = OpenAIProvider()
        assert p._default_model == "gpt-4o"

    @patch("openai.OpenAI")
    @patch(
        "akande.config.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    def test_error_logging_on_failure(self, mock_cls):
        from akande.providers.openai_provider import (
            OpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.side_effect = (
            RuntimeError("api down")
        )
        p = OpenAIProvider()
        with pytest.raises(RuntimeError, match="api down"):
            p.generate_response_sync(
                "hi", "sys", "gpt-4"
            )


# ────────────────────────────────────────────────────────────
# OpenAI-Compatible Providers (Ollama, Groq, LM Studio)
# ────────────────────────────────────────────────────────────


class TestOllamaProvider:

    @patch("openai.OpenAI")
    def test_provider_name(self, mock_cls):
        from akande.providers.ollama_provider import (
            OllamaProvider,
        )
        p = OllamaProvider()
        assert p.provider_name == "ollama"

    @patch("openai.OpenAI")
    def test_default_base_url(self, mock_cls):
        from akande.providers.ollama_provider import (
            OllamaProvider,
        )
        p = OllamaProvider()
        assert p._base_url == "http://localhost:11434/v1"

    @patch.dict(
        "os.environ",
        {"OLLAMA_HOST": "http://remote:11434"},
    )
    @patch("openai.OpenAI")
    def test_custom_host(self, mock_cls):
        from akande.providers.ollama_provider import (
            OllamaProvider,
        )
        p = OllamaProvider()
        assert p._base_url == "http://remote:11434/v1"

    @patch("openai.OpenAI")
    def test_sync_call(self, mock_cls):
        from akande.providers.ollama_provider import (
            OllamaProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = OllamaProvider()
        result = p.generate_response_sync(
            "hi", "sys", "llama3"
        )
        assert result is not None


class TestGroqProvider:

    @patch("openai.OpenAI")
    @patch.dict(
        "os.environ", {"GROQ_API_KEY": "gsk_testkey123"}
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.groq_provider import (
            GroqProvider,
        )
        p = GroqProvider()
        assert p.provider_name == "groq"

    @patch("openai.OpenAI")
    @patch.dict(
        "os.environ", {"GROQ_API_KEY": "gsk_testkey123"}
    )
    def test_base_url(self, mock_cls):
        from akande.providers.groq_provider import (
            GroqProvider,
        )
        p = GroqProvider()
        assert "groq.com" in p._base_url

    @patch("openai.OpenAI")
    @patch.dict(
        "os.environ", {"GROQ_API_KEY": "gsk_testkey123"}
    )
    def test_sync_call(self, mock_cls):
        from akande.providers.groq_provider import (
            GroqProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = GroqProvider()
        result = p.generate_response_sync(
            "hi", "sys", "llama3-8b-8192"
        )
        assert result is not None

    def test_missing_api_key_raises(self):
        from akande.providers.groq_provider import (
            GroqProvider,
        )
        with patch.dict("os.environ", {}, clear=True):
            with patch.dict(
                "os.environ",
                {"GROQ_API_KEY": ""},
            ):
                with pytest.raises(ValueError, match="API key"):
                    GroqProvider()


class TestLMStudioProvider:

    @patch("openai.OpenAI")
    def test_provider_name(self, mock_cls):
        from akande.providers.lmstudio_provider import (
            LMStudioProvider,
        )
        p = LMStudioProvider()
        assert p.provider_name == "lmstudio"

    @patch("openai.OpenAI")
    def test_default_base_url(self, mock_cls):
        from akande.providers.lmstudio_provider import (
            LMStudioProvider,
        )
        p = LMStudioProvider()
        assert p._base_url == "http://localhost:1234/v1"

    @patch.dict(
        "os.environ",
        {"LMSTUDIO_HOST": "http://myhost:5555"},
    )
    @patch("openai.OpenAI")
    def test_custom_host(self, mock_cls):
        from akande.providers.lmstudio_provider import (
            LMStudioProvider,
        )
        p = LMStudioProvider()
        assert p._base_url == "http://myhost:5555/v1"


# ────────────────────────────────────────────────────────────
# Azure OpenAI Provider
# ────────────────────────────────────────────────────────────


class TestAzureOpenAIProvider:

    @patch("openai.AzureOpenAI")
    @patch.dict(
        "os.environ",
        {
            "AZURE_OPENAI_API_KEY": "test-key",
            "AZURE_OPENAI_ENDPOINT": "https://test.azure.com",
        },
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.azure_openai_provider import (
            AzureOpenAIProvider,
        )
        p = AzureOpenAIProvider()
        assert p.provider_name == "azure_openai"

    @patch.dict(
        "os.environ",
        {
            "AZURE_OPENAI_API_KEY": "test-key",
            "AZURE_OPENAI_ENDPOINT": "https://test.openai.azure.com",
            "AZURE_OPENAI_API_VERSION": "2024-02-01",
        },
    )
    @patch("openai.AzureOpenAI")
    def test_reads_env_vars(self, mock_cls):
        from akande.providers.azure_openai_provider import (
            AzureOpenAIProvider,
        )
        p = AzureOpenAIProvider()
        assert p._api_key == "test-key"

    @patch("openai.AzureOpenAI")
    @patch.dict(
        "os.environ",
        {
            "AZURE_OPENAI_API_KEY": "test-key",
            "AZURE_OPENAI_ENDPOINT": "https://test.azure.com",
        },
    )
    def test_sync_call(self, mock_cls):
        from akande.providers.azure_openai_provider import (
            AzureOpenAIProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.chat.completions.create.return_value = (
            MagicMock()
        )
        p = AzureOpenAIProvider()
        result = p.generate_response_sync(
            "hi", "sys", "gpt-35-turbo"
        )
        assert result is not None

    @patch.dict(
        "os.environ",
        {"AZURE_OPENAI_ENDPOINT": "https://test.azure.com"},
        clear=False,
    )
    def test_missing_api_key_raises(self):
        from akande.providers.azure_openai_provider import (
            AzureOpenAIProvider,
        )
        import os
        os.environ.pop("AZURE_OPENAI_API_KEY", None)
        with pytest.raises(
            ValueError, match="AZURE_OPENAI_API_KEY"
        ):
            AzureOpenAIProvider()

    @patch.dict(
        "os.environ",
        {"AZURE_OPENAI_API_KEY": "test-key"},
        clear=False,
    )
    def test_missing_endpoint_raises(self):
        from akande.providers.azure_openai_provider import (
            AzureOpenAIProvider,
        )
        import os
        os.environ.pop("AZURE_OPENAI_ENDPOINT", None)
        with pytest.raises(
            ValueError, match="AZURE_OPENAI_ENDPOINT"
        ):
            AzureOpenAIProvider()


# ────────────────────────────────────────────────────────────
# Anthropic Provider
# ────────────────────────────────────────────────────────────


class TestAnthropicProvider:

    @patch("anthropic.Anthropic")
    @patch.dict(
        "os.environ",
        {"ANTHROPIC_API_KEY": "sk-ant-test123"},
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        p = AnthropicProvider()
        assert p.provider_name == "anthropic"

    @patch("anthropic.Anthropic")
    @patch.dict(
        "os.environ",
        {"ANTHROPIC_API_KEY": "sk-ant-test123"},
    )
    def test_sync_call_returns_provider_response(
        self, mock_cls
    ):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_block = MagicMock()
        mock_block.text = "hello response"
        mock_response = MagicMock()
        mock_response.content = [mock_block]
        mock_client.messages.create.return_value = (
            mock_response
        )

        p = AnthropicProvider()
        result = p.generate_response_sync(
            "hi", "sys", "claude-3-haiku-20240307"
        )
        assert result.choices[0].message.content == (
            "hello response"
        )

    @patch("anthropic.Anthropic")
    @patch.dict(
        "os.environ",
        {"ANTHROPIC_API_KEY": "sk-ant-test123"},
    )
    def test_async_call(self, mock_cls):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_block = MagicMock()
        mock_block.text = "async result"
        mock_response = MagicMock()
        mock_response.content = [mock_block]
        mock_client.messages.create.return_value = (
            mock_response
        )

        p = AnthropicProvider()
        result = asyncio.run(
            p.generate_response(
                "hi", "sys", "claude-3-haiku-20240307"
            )
        )
        assert result.choices[0].message.content == (
            "async result"
        )

    @patch("anthropic.Anthropic")
    @patch.dict(
        "os.environ",
        {"ANTHROPIC_API_KEY": "sk-ant-test123"},
    )
    def test_passes_system_as_param(self, mock_cls):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_block = MagicMock()
        mock_block.text = ""
        mock_response = MagicMock()
        mock_response.content = [mock_block]
        mock_client.messages.create.return_value = (
            mock_response
        )

        p = AnthropicProvider()
        p.generate_response_sync(
            "user msg", "system msg",
            "claude-3-haiku-20240307",
        )
        args = mock_client.messages.create.call_args
        assert args.kwargs.get("system") == "system msg"

    def test_missing_api_key_raises(self):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        with patch.dict(
            "os.environ", {"ANTHROPIC_API_KEY": ""}
        ):
            with pytest.raises(
                ValueError, match="ANTHROPIC_API_KEY"
            ):
                AnthropicProvider()

    @patch("anthropic.Anthropic")
    @patch.dict(
        "os.environ",
        {"ANTHROPIC_API_KEY": "sk-ant-test123"},
    )
    def test_error_logging_on_failure(self, mock_cls):
        from akande.providers.anthropic_provider import (
            AnthropicProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client
        mock_client.messages.create.side_effect = (
            RuntimeError("api down")
        )
        p = AnthropicProvider()
        with pytest.raises(RuntimeError, match="api down"):
            p.generate_response_sync(
                "hi", "sys", "claude-3-haiku-20240307"
            )


# ────────────────────────────────────────────────────────────
# Google Gemini Provider
# ────────────────────────────────────────────────────────────


class TestGoogleProvider:

    @patch("google.generativeai.configure")
    @patch("google.generativeai.GenerativeModel")
    @patch.dict(
        "os.environ",
        {"GOOGLE_API_KEY": "test-google-key"},
    )
    def test_provider_name(self, mock_model, mock_configure):
        from akande.providers.google_provider import (
            GoogleProvider,
        )
        p = GoogleProvider()
        assert p.provider_name == "google"

    @patch("google.generativeai.configure")
    @patch("google.generativeai.GenerativeModel")
    @patch.dict(
        "os.environ",
        {"GOOGLE_API_KEY": "test-google-key"},
    )
    def test_sync_call(self, mock_model_cls, mock_configure):
        from akande.providers.google_provider import (
            GoogleProvider,
        )
        mock_model = MagicMock()
        mock_model_cls.return_value = mock_model
        mock_response = MagicMock()
        mock_response.text = "gemini says hi"
        mock_model.generate_content.return_value = (
            mock_response
        )

        p = GoogleProvider()
        result = p.generate_response_sync(
            "hi", "sys", "gemini-pro"
        )
        assert result.choices[0].message.content == (
            "gemini says hi"
        )

    def test_missing_api_key_raises(self):
        from akande.providers.google_provider import (
            GoogleProvider,
        )
        with patch("google.generativeai.configure"):
            with patch.dict(
                "os.environ", {"GOOGLE_API_KEY": ""}
            ):
                with pytest.raises(
                    ValueError, match="GOOGLE_API_KEY"
                ):
                    GoogleProvider()


# ────────────────────────────────────────────────────────────
# Mistral Provider
# ────────────────────────────────────────────────────────────


class TestMistralProvider:

    @patch("mistralai.Mistral")
    @patch.dict(
        "os.environ",
        {"MISTRAL_API_KEY": "test-mistral-key"},
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.mistral_provider import (
            MistralProvider,
        )
        p = MistralProvider()
        assert p.provider_name == "mistral"

    @patch("mistralai.Mistral")
    @patch.dict(
        "os.environ",
        {"MISTRAL_API_KEY": "test-mistral-key"},
    )
    def test_sync_call(self, mock_cls):
        from akande.providers.mistral_provider import (
            MistralProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_choice = MagicMock()
        mock_choice.message.content = "mistral response"
        mock_response = MagicMock()
        mock_response.choices = [mock_choice]
        mock_client.chat.complete.return_value = mock_response

        p = MistralProvider()
        result = p.generate_response_sync(
            "hi", "sys", "mistral-small-latest"
        )
        assert result.choices[0].message.content == (
            "mistral response"
        )

    def test_missing_api_key_raises(self):
        from akande.providers.mistral_provider import (
            MistralProvider,
        )
        with patch.dict(
            "os.environ", {"MISTRAL_API_KEY": ""}
        ):
            with pytest.raises(
                ValueError, match="MISTRAL_API_KEY"
            ):
                MistralProvider()


# ────────────────────────────────────────────────────────────
# Cohere Provider
# ────────────────────────────────────────────────────────────


class TestCohereProvider:

    @patch("cohere.ClientV2")
    @patch.dict(
        "os.environ",
        {"COHERE_API_KEY": "test-cohere-key"},
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.cohere_provider import (
            CohereProvider,
        )
        p = CohereProvider()
        assert p.provider_name == "cohere"

    @patch("cohere.ClientV2")
    @patch.dict(
        "os.environ",
        {"COHERE_API_KEY": "test-cohere-key"},
    )
    def test_sync_call(self, mock_cls):
        from akande.providers.cohere_provider import (
            CohereProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_text_block = MagicMock()
        mock_text_block.text = "cohere response"
        mock_msg = MagicMock()
        mock_msg.content = [mock_text_block]
        mock_response = MagicMock()
        mock_response.message = mock_msg
        mock_client.chat.return_value = mock_response

        p = CohereProvider()
        result = p.generate_response_sync(
            "hi", "sys", "command-r"
        )
        assert result.choices[0].message.content == (
            "cohere response"
        )

    def test_missing_api_key_raises(self):
        from akande.providers.cohere_provider import (
            CohereProvider,
        )
        with patch.dict(
            "os.environ", {"COHERE_API_KEY": ""}
        ):
            with pytest.raises(
                ValueError, match="COHERE_API_KEY"
            ):
                CohereProvider()


# ────────────────────────────────────────────────────────────
# Hugging Face Provider
# ────────────────────────────────────────────────────────────


class TestHuggingFaceProvider:

    @patch("huggingface_hub.InferenceClient")
    @patch.dict(
        "os.environ",
        {"HUGGINGFACE_API_KEY": "hf_testtoken123"},
    )
    def test_provider_name(self, mock_cls):
        from akande.providers.huggingface_provider import (
            HuggingFaceProvider,
        )
        p = HuggingFaceProvider()
        assert p.provider_name == "huggingface"

    @patch("huggingface_hub.InferenceClient")
    @patch.dict(
        "os.environ",
        {"HUGGINGFACE_API_KEY": "hf_testtoken123"},
    )
    def test_sync_call(self, mock_cls):
        from akande.providers.huggingface_provider import (
            HuggingFaceProvider,
        )
        mock_client = MagicMock()
        mock_cls.return_value = mock_client

        mock_choice = MagicMock()
        mock_choice.message.content = "hf response"
        mock_response = MagicMock()
        mock_response.choices = [mock_choice]
        mock_client.chat_completion.return_value = (
            mock_response
        )

        p = HuggingFaceProvider()
        result = p.generate_response_sync(
            "hi", "sys",
            "mistralai/Mistral-7B-Instruct-v0.2",
        )
        assert result.choices[0].message.content == (
            "hf response"
        )

    def test_missing_api_key_raises(self):
        from akande.providers.huggingface_provider import (
            HuggingFaceProvider,
        )
        with patch.dict(
            "os.environ", {"HUGGINGFACE_API_KEY": ""}
        ):
            with pytest.raises(
                ValueError, match="HUGGINGFACE_API_KEY"
            ):
                HuggingFaceProvider()


# ────────────────────────────────────────────────────────────
# Adapter Compliance: all providers share the LLMProvider ABC
# ────────────────────────────────────────────────────────────


class TestAdapterCompliance:
    """Verify every registered provider is a LLMProvider."""

    def test_all_registered_are_llm_providers(self):
        for name in _registry.available:
            cls = _registry._resolve_class(name)
            assert issubclass(cls, LLMProvider), (
                f"{name} ({cls}) is not a LLMProvider subclass"
            )

    def test_all_have_provider_name_property(self):
        for name in _registry.available:
            cls = _registry._resolve_class(name)
            assert hasattr(cls, "provider_name")

    def test_all_have_generate_response(self):
        for name in _registry.available:
            cls = _registry._resolve_class(name)
            assert hasattr(cls, "generate_response")

    def test_all_have_generate_response_sync(self):
        for name in _registry.available:
            cls = _registry._resolve_class(name)
            assert hasattr(cls, "generate_response_sync")


# ────────────────────────────────────────────────────────────
# Import error handling for native SDK providers
# ────────────────────────────────────────────────────────────


class TestImportErrors:
    """Verify clear ImportError for missing optional SDKs."""

    def test_anthropic_import_error(self):
        import sys
        saved = sys.modules.get("anthropic")
        sys.modules["anthropic"] = None
        try:
            from akande.providers.anthropic_provider import (
                AnthropicProvider,
            )
            with pytest.raises(ImportError, match="anthropic"):
                AnthropicProvider()
        finally:
            if saved is not None:
                sys.modules["anthropic"] = saved
            else:
                sys.modules.pop("anthropic", None)

    def test_google_import_error(self):
        import sys
        saved = sys.modules.get("google.generativeai")
        saved_google = sys.modules.get("google")
        sys.modules["google.generativeai"] = None
        sys.modules["google"] = None
        try:
            from akande.providers.google_provider import (
                GoogleProvider,
            )
            with pytest.raises(ImportError):
                GoogleProvider()
        finally:
            if saved is not None:
                sys.modules["google.generativeai"] = saved
            else:
                sys.modules.pop("google.generativeai", None)
            if saved_google is not None:
                sys.modules["google"] = saved_google
            else:
                sys.modules.pop("google", None)

    def test_mistral_import_error(self):
        import sys
        saved = sys.modules.get("mistralai")
        sys.modules["mistralai"] = None
        try:
            from akande.providers.mistral_provider import (
                MistralProvider,
            )
            with pytest.raises(ImportError, match="mistralai"):
                MistralProvider()
        finally:
            if saved is not None:
                sys.modules["mistralai"] = saved
            else:
                sys.modules.pop("mistralai", None)

    def test_cohere_import_error(self):
        import sys
        saved = sys.modules.get("cohere")
        sys.modules["cohere"] = None
        try:
            from akande.providers.cohere_provider import (
                CohereProvider,
            )
            with pytest.raises(ImportError, match="cohere"):
                CohereProvider()
        finally:
            if saved is not None:
                sys.modules["cohere"] = saved
            else:
                sys.modules.pop("cohere", None)

    def test_huggingface_import_error(self):
        import sys
        saved = sys.modules.get("huggingface_hub")
        sys.modules["huggingface_hub"] = None
        try:
            from akande.providers.huggingface_provider import (
                HuggingFaceProvider,
            )
            with pytest.raises(
                ImportError, match="huggingface_hub"
            ):
                HuggingFaceProvider()
        finally:
            if saved is not None:
                sys.modules["huggingface_hub"] = saved
            else:
                sys.modules.pop("huggingface_hub", None)



import asyncio
from unittest.mock import MagicMock, patch

import pytest

from akande.services import (
    OpenAIService,
    OpenAIImpl,
    SYSTEM_PROMPT,
)


class TestSystemPrompt:
    def test_system_prompt_contains_akande(self):
        assert "Àkàndé" in SYSTEM_PROMPT

    def test_system_prompt_contains_structure(self):
        assert "Overview" in SYSTEM_PROMPT
        assert "Solution" in SYSTEM_PROMPT
        assert "Conclusion" in SYSTEM_PROMPT


class TestOpenAIServiceABC:
    def test_cannot_instantiate_abc(self):
        with pytest.raises(TypeError):
            OpenAIService()


class TestOpenAIImpl:
    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_init_passes_timeout(self, mock_openai_cls):
        OpenAIImpl()
        mock_openai_cls.assert_called_once()
        call_kwargs = mock_openai_cls.call_args.kwargs
        assert "timeout" in call_kwargs

    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_generate_response_success(self, mock_openai_cls):
        mock_client = MagicMock()
        mock_openai_cls.return_value = mock_client

        mock_response = MagicMock()
        mock_response.choices = [
            MagicMock(message=MagicMock(content="Test response"))
        ]
        mock_client.chat.completions.create.return_value = (
            mock_response
        )

        service = OpenAIImpl()
        result = asyncio.run(
            service.generate_response(
                "Hello",
                SYSTEM_PROMPT,
                "gpt-3.5-turbo",
                {},
            )
        )

        assert result == mock_response
        mock_client.chat.completions.create.assert_called_once()

    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_generate_response_uses_system_role(
        self, mock_openai_cls
    ):
        mock_client = MagicMock()
        mock_openai_cls.return_value = mock_client

        mock_response = MagicMock()
        mock_client.chat.completions.create.return_value = (
            mock_response
        )

        service = OpenAIImpl()
        asyncio.run(
            service.generate_response(
                "What is AI?",
                SYSTEM_PROMPT,
                "gpt-4",
                {},
            )
        )

        call_args = mock_client.chat.completions.create.call_args
        messages = call_args.kwargs.get("messages", [])
        assert messages[0]["role"] == "system"
        assert messages[1]["role"] == "user"
        assert "Àkàndé" in messages[0]["content"]
        assert "What is AI?" in messages[1]["content"]

    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_generate_response_propagates_exception(
        self, mock_openai_cls
    ):
        mock_client = MagicMock()
        mock_openai_cls.return_value = mock_client
        mock_client.chat.completions.create.side_effect = (
            Exception("API error")
        )

        service = OpenAIImpl()
        with pytest.raises(Exception, match="API error"):
            asyncio.run(
                service.generate_response(
                    "Hello",
                    SYSTEM_PROMPT,
                    "gpt-3.5-turbo",
                    {},
                )
            )

    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_generate_response_sync(self, mock_openai_cls):
        mock_client = MagicMock()
        mock_openai_cls.return_value = mock_client

        mock_response = MagicMock()
        mock_response.choices = [
            MagicMock(message=MagicMock(content="Sync response"))
        ]
        mock_client.chat.completions.create.return_value = (
            mock_response
        )

        service = OpenAIImpl()
        result = service.generate_response_sync(
            "Hello", SYSTEM_PROMPT, "gpt-3.5-turbo", {}
        )

        assert result == mock_response

    @patch(
        "akande.services.OPENAI_API_KEY",
        "sk-test-1234567890abcdef",
    )
    @patch("akande.services.openai.OpenAI")
    def test_sync_uses_system_and_user_roles(
        self, mock_openai_cls
    ):
        mock_client = MagicMock()
        mock_openai_cls.return_value = mock_client

        mock_response = MagicMock()
        mock_client.chat.completions.create.return_value = (
            mock_response
        )

        service = OpenAIImpl()
        service.generate_response_sync(
            "Test", SYSTEM_PROMPT, "gpt-4", {}
        )

        call_args = mock_client.chat.completions.create.call_args
        messages = call_args.kwargs.get("messages", [])
        assert messages[0]["role"] == "system"
        assert messages[1]["role"] == "user"



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional


class LLMProvider(ABC):
    """Abstract base class for LLM providers.

    All providers must implement both async and sync response
    generation methods. Implementations should handle their own
    credential loading from environment variables.
    """

    @property
    @abstractmethod
    def provider_name(self) -> str:
        """Return the provider name for logging."""
        ...

    @abstractmethod
    async def generate_response(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Generate a response asynchronously.

        Parameters
        ----------
        user_prompt : str
            The user's input prompt.
        system_prompt : str
            The system prompt for context.
        model : str
            The model identifier.
        params : dict, optional
            Additional provider-specific parameters.

        Returns
        -------
        Any
            The provider's response object.
        """
        ...

    @abstractmethod
    def generate_response_sync(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Generate a response synchronously.

        Parameters
        ----------
        user_prompt : str
            The user's input prompt.
        system_prompt : str
            The system prompt for context.
        model : str
            The model identifier.
        params : dict, optional
            Additional provider-specific parameters.

        Returns
        -------
        Any
            The provider's response object.
        """
        ...



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import csv
from typing import Optional
from xml.sax.saxutils import escape as xml_escape
from reportlab.lib.pagesizes import letter
from reportlab.platypus import (
    SimpleDocTemplate,
    Paragraph,
    Spacer,
    Image,
)
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.enums import TA_LEFT
from pathlib import Path
import datetime
import logging
import re

# Module-level cached styles for PDF generation
_styles = getSampleStyleSheet()

_list_item_style = ParagraphStyle(
    "listItem",
    parent=_styles["BodyText"],
    fontSize=12,
    leading=14,
    spaceBefore=0,
    spaceAfter=6,
    leftIndent=10,
    firstLineIndent=-10,
)

_heading1_style = _styles["Heading1"]
_heading1_style.fontName = "Helvetica-Bold"
_heading1_style.fontSize = 14
_heading1_style.leading = 16
_heading1_style.alignment = TA_LEFT

_heading2_style = _styles["Heading2"]
_heading2_style.fontName = "Helvetica-Bold"
_heading2_style.fontSize = 12
_heading2_style.leading = 14
_heading2_style.alignment = TA_LEFT

_paragraph_style = _styles["BodyText"]
_paragraph_style.fontName = "Helvetica"
_paragraph_style.fontSize = 12
_paragraph_style.leading = 14
_paragraph_style.alignment = TA_LEFT


def get_output_directory() -> Path:
    """
    Get or create the date-stamped output directory.

    Returns
    -------
    Path
        The path to the output directory for today's date.
    """
    date_str = datetime.datetime.now().strftime("%Y-%m-%d")
    directory_path = Path(date_str)
    directory_path.mkdir(parents=True, exist_ok=True)
    return directory_path


def get_output_filename(extension: str) -> str:
    """
    Generate a timestamped output filename with seconds precision.

    Parameters
    ----------
    extension : str
        The file extension (e.g., '.pdf', '.csv', '.wav', '.log').

    Returns
    -------
    str
        The generated filename.
    """
    return (
        datetime.datetime.now().strftime(
            "%Y-%m-%d-%H-%M-%S-Akande"
        )
        + extension
    )


def validate_api_key(api_key: Optional[str]) -> bool:
    """
    Validates the format of an OpenAI API key.

    Parameters
    ----------
    api_key : Optional[str]
        The API key to validate.

    Returns
    -------
    bool
        True if the API key format is valid, False otherwise.
    """
    if api_key is None or len(api_key) < 20:
        return False
    valid_prefixes = ("sk-", "sk-proj-", "sk-org-")
    return api_key.startswith(valid_prefixes)


def generate_pdf(question: str, response: str) -> None:
    """
    Generates a PDF document containing a question and response.

    Parameters
    ----------
    question : str
        The question to be included in the PDF.
    response : str
        The response to the question.
    """
    try:
        directory_path = get_output_directory()
        filename = get_output_filename(".pdf")
        file_path = directory_path / filename

        doc = SimpleDocTemplate(str(file_path), pagesize=letter)
        flowables = []

        # Optional: Add a logo at the top if the file exists
        logo_path = Path("./512x512.png")
        if logo_path.exists():
            logo = Image(str(logo_path), width=48, height=48)
            logo.hAlign = "RIGHT"
            logo.preserveAspectRatio = True
            flowables.append(logo)
            flowables.append(Spacer(1, 12))

        # Escape user input to prevent ReportLab markup injection
        safe_question = xml_escape(question.title())
        flowables.append(
            Paragraph(safe_question, _heading1_style)
        )
        flowables.append(Spacer(1, 6))

        # Process and format the response content
        paragraphs = response.split("\n")
        for para in paragraphs:
            safe_para = xml_escape(para)
            if para.startswith(
                (
                    "Overview",
                    "Solution",
                    "Conclusion",
                    "Recommendations",
                )
            ):
                flowables.append(
                    Paragraph(safe_para, _heading2_style)
                )
                flowables.append(Spacer(1, 6))
            elif re.match(r"^-?\d", para):
                formatted_text = (
                    "- " + safe_para
                    if not safe_para.startswith("-")
                    else safe_para
                )
                flowables.append(
                    Paragraph(formatted_text, _list_item_style)
                )
                flowables.append(Spacer(1, 6))
            else:
                flowables.append(
                    Paragraph(safe_para, _paragraph_style)
                )
                flowables.append(Spacer(1, 6))

        doc.build(flowables)
        logging.info(
            "PDF generated",
            extra={
                "event": "Export:PDFGenerated",
                "extra_data": {"file_path": str(file_path)},
            },
        )
    except Exception as e:
        logging.error(
            f"PDF generation failed: {type(e).__name__}: {e}",
            exc_info=True,
            extra={"event": "Export:PDFFailed"},
        )


def generate_csv(question: str, response: str) -> None:
    """
    Generates a CSV document containing a question and response.

    Parameters
    ----------
    question : str
        The question to be included in the CSV.
    response : str
        The response to the question.
    """
    try:
        directory_path = get_output_directory()
        filename = get_output_filename(".csv")
        file_path = directory_path / filename

        with open(
            file_path, mode="w", newline="", encoding="utf-8"
        ) as file:
            csv_writer = csv.writer(file)
            csv_writer.writerow(["Question", "Response"])
            csv_writer.writerow([question, response])

        logging.info(
            "CSV generated",
            extra={
                "event": "Export:CSVGenerated",
                "extra_data": {"file_path": str(file_path)},
            },
        )
    except Exception as e:
        logging.error(
            f"CSV generation failed: {type(e).__name__}: {e}",
            exc_info=True,
            extra={"event": "Export:CSVFailed"},
        )



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyncio
import logging
import time
from typing import Any, Dict, Optional

from .base import LLMProvider

# Providers that use a local server and do not need a
# real API key.
_LOCAL_PROVIDERS = frozenset({"ollama", "lmstudio"})


class OpenAICompatProvider(LLMProvider):
    """Base class for providers using an OpenAI-compatible API.

    Subclasses only need to set _provider_name, _base_url,
    _api_key, and _default_model in their __init__.
    """

    _provider_name: str = ""
    _base_url: str = ""
    _api_key: str = ""
    _default_model: str = ""

    def _init_client(self):
        """Initialise the OpenAI-compatible client."""
        import openai
        from akande.config import API_CALL_TIMEOUT

        api_key = self._api_key
        if (
            not api_key
            and self._provider_name in _LOCAL_PROVIDERS
        ):
            api_key = self._provider_name
        elif not api_key:
            raise ValueError(
                f"API key is required for the "
                f"'{self._provider_name}' provider. "
                f"Set the appropriate environment variable."
            )

        kwargs: Dict[str, Any] = {
            "api_key": api_key,
            "timeout": API_CALL_TIMEOUT,
        }
        if self._base_url:
            kwargs["base_url"] = self._base_url
        self.client = openai.OpenAI(**kwargs)

    @property
    def provider_name(self) -> str:
        return self._provider_name

    async def generate_response(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if not params:
            params = {}
        model = model or self._default_model

        logging.info(
            "LLM request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": self._provider_name,
                    "model": model,
                },
            },
        )
        start = time.time()
        try:
            loop = asyncio.get_running_loop()
            response = await loop.run_in_executor(
                None,
                lambda: self.client.chat.completions.create(
                    model=model,
                    messages=[
                        {
                            "role": "system",
                            "content": system_prompt,
                        },
                        {
                            "role": "user",
                            "content": user_prompt,
                        },
                    ],
                    **params,
                ),
            )
        except Exception:
            latency = (time.time() - start) * 1000
            logging.error(
                "LLM request failed",
                exc_info=True,
                extra={
                    "event": "LLM:RequestFailed",
                    "extra_data": {
                        "provider": self._provider_name,
                        "model": model,
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            raise
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": self._provider_name,
                    "model": model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response

    def generate_response_sync(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if not params:
            params = {}
        model = model or self._default_model

        logging.info(
            "LLM sync request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": self._provider_name,
                    "model": model,
                },
            },
        )
        start = time.time()
        try:
            response = self.client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "system",
                        "content": system_prompt,
                    },
                    {
                        "role": "user",
                        "content": user_prompt,
                    },
                ],
                **params,
            )
        except Exception:
            latency = (time.time() - start) * 1000
            logging.error(
                "LLM sync request failed",
                exc_info=True,
                extra={
                    "event": "LLM:RequestFailed",
                    "extra_data": {
                        "provider": self._provider_name,
                        "model": model,
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            raise
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM sync response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": self._provider_name,
                    "model": model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from dotenv import load_dotenv

# Loads environment variables from a .env file into the environment.
load_dotenv()

# Configuration

# Number of seconds allowed per API call before timing out.
API_CALL_TIMEOUT = 90

# Gets OpenAI API key from the environment variables loaded by load_dotenv().
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

# Gets the name of default OpenAI model to use from the environment variables
OPENAI_DEFAULT_MODEL = os.getenv("OPENAI_DEFAULT_MODEL")

# LLM provider to use (default: "openai"). See akande/providers/ for options.
# Available: openai, anthropic, google, ollama, azure_openai, mistral,
#            cohere, huggingface, groq, lmstudio
LLM_PROVIDER = os.getenv("LLM_PROVIDER", "openai")



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import sys
import traceback
from datetime import datetime, timezone


class JSONFormatter(logging.Formatter):
    """Structured JSON log formatter with mandatory context fields."""

    def __init__(self, service: str = "akande"):
        super().__init__()
        self.service = service

    def format(self, record: logging.LogRecord) -> str:
        log_entry = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "level": record.levelname,
            "logger": record.name,
            "service": self.service,
            "message": record.getMessage(),
        }
        # Include correlation_id if set on the record
        correlation_id = getattr(record, "correlation_id", None)
        if correlation_id:
            log_entry["correlation_id"] = correlation_id

        # Include event name if set on the record
        event = getattr(record, "event", None)
        if event:
            log_entry["event"] = event

        # Include extra data if set on the record
        extra_data = getattr(record, "extra_data", None)
        if extra_data and isinstance(extra_data, dict):
            log_entry["data"] = extra_data

        if record.exc_info and record.exc_info[0] is not None:
            log_entry["exception"] = {
                "type": record.exc_info[0].__name__,
                "message": str(record.exc_info[1]),
                "traceback": traceback.format_exception(
                    *record.exc_info
                ),
            }
        return json.dumps(log_entry, default=str)


def basic_config(filename: str, level: int, log_format: str) -> None:
    """
    Configure logging with both file and console handlers.

    Uses JSON structured logging for file output and a human-readable
    format for console output.

    :param filename: The name of the log file.
    :param level: The logging level.
    :param log_format: The format of the log messages (used for
        console output).
    :return: None
    """
    root = logging.getLogger()
    root.setLevel(level)

    # Remove any existing handlers to avoid duplicates
    for handler in root.handlers[:]:
        root.removeHandler(handler)

    # File handler uses JSON structured logging
    json_formatter = JSONFormatter(service="akande")
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(json_formatter)
    root.addHandler(file_handler)

    # Console handler uses human-readable format
    console_formatter = logging.Formatter(log_format)
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(console_formatter)
    root.addHandler(console_handler)



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC, abstractmethod
import asyncio
import logging
import time
from typing import Any, Dict, Optional
import openai
from .config import (
    OPENAI_API_KEY,
    OPENAI_DEFAULT_MODEL,
    API_CALL_TIMEOUT,
)

SYSTEM_PROMPT = """
As "\u00c0k\u00e0nd\u00e9," an AI assistant, your mission is to support users
providing accurate information on various topics, condensed into a
concise yet comprehensive briefing. Respond only in grammatically
correct British English using proper spelling and local terminology.

Adhere to a 150-word structure:

Overview
- Briefly introduce the topic and frame the key question(s) to be
addressed, highlighting relevance to the user.

Solution
- Offer an actionable response using bullet points for clarity.
- Outline technical solutions or conceptual recommendations.

Conclusion
- Concisely summarize 2-3 most important conclusions or next steps
for the user.

Recommendations
- Provide helpful recommendations based on the information presented.

Use straightforward language suitable for a middle-school audience.
Avoid profanity or potentially insensitive language. Focus on
delivering value by prioritising essential information relevant to the
user's needs within 150 words.
""".strip()


class OpenAIService(ABC):
    """Base class for OpenAI services.

    Retained for backward compatibility. New code should use
    akande.providers.LLMProvider instead.
    """

    @abstractmethod
    async def generate_response(
        self,
        prompt: str,
        system_prompt: str = "",
        model: str = "",
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        pass


class OpenAIImpl(OpenAIService):
    """OpenAI API client implementation.

    Retained for backward compatibility. Delegates to
    akande.providers.openai_provider.OpenAIProvider internally.
    """

    def __init__(self):
        self.client = openai.OpenAI(
            api_key=OPENAI_API_KEY,
            timeout=API_CALL_TIMEOUT,
        )

    async def generate_response(
        self,
        user_prompt: str,
        system_prompt: str = "",
        model: str = "",
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if not params:
            params = {}
        model = model or OPENAI_DEFAULT_MODEL
        system_prompt = system_prompt or SYSTEM_PROMPT

        logging.info(
            "LLM request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": "openai",
                    "model": model,
                },
            },
        )
        start = time.time()
        loop = asyncio.get_running_loop()
        response = await loop.run_in_executor(
            None,
            lambda: self.client.chat.completions.create(
                model=model,
                messages=[
                    {
                        "role": "system",
                        "content": system_prompt,
                    },
                    {"role": "user", "content": user_prompt},
                ],
                **params,
            ),
        )
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": "openai",
                    "model": model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response

    def generate_response_sync(
        self,
        user_prompt: str,
        system_prompt: str = "",
        model: str = "",
        params: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if not params:
            params = {}
        model = model or OPENAI_DEFAULT_MODEL
        system_prompt = system_prompt or SYSTEM_PROMPT

        logging.info(
            "LLM sync request sent",
            extra={
                "event": "LLM:RequestSent",
                "extra_data": {
                    "provider": "openai",
                    "model": model,
                },
            },
        )
        start = time.time()
        response = self.client.chat.completions.create(
            model=model,
            messages=[
                {
                    "role": "system",
                    "content": system_prompt,
                },
                {"role": "user", "content": user_prompt},
            ],
            **params,
        )
        latency = (time.time() - start) * 1000
        logging.info(
            "LLM sync response received",
            extra={
                "event": "LLM:ResponseReceived",
                "extra_data": {
                    "provider": "openai",
                    "model": model,
                    "latency_ms": round(latency, 2),
                },
            },
        )
        return response



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import importlib
import logging
from typing import Dict, List, Optional, Type

from .base import LLMProvider

DEFAULT_PROVIDER = "openai"

# Lazy import mapping: provider name -> (module_path, class_name)
_PROVIDER_MAP: Dict[str, tuple] = {
    "openai": (
        ".openai_provider",
        "OpenAIProvider",
    ),
    "azure_openai": (
        ".azure_openai_provider",
        "AzureOpenAIProvider",
    ),
    "ollama": (
        ".ollama_provider",
        "OllamaProvider",
    ),
    "lmstudio": (
        ".lmstudio_provider",
        "LMStudioProvider",
    ),
    "groq": (
        ".groq_provider",
        "GroqProvider",
    ),
    "anthropic": (
        ".anthropic_provider",
        "AnthropicProvider",
    ),
    "google": (
        ".google_provider",
        "GoogleProvider",
    ),
    "mistral": (
        ".mistral_provider",
        "MistralProvider",
    ),
    "cohere": (
        ".cohere_provider",
        "CohereProvider",
    ),
    "huggingface": (
        ".huggingface_provider",
        "HuggingFaceProvider",
    ),
}


class ProviderRegistry:
    """Registry for LLM provider implementations.

    Providers are registered lazily by name. The provider module
    is only imported when ``create()`` is first called for that
    name, and the resulting instance is cached for reuse.
    """

    def __init__(
        self,
        lazy_map: Optional[Dict[str, tuple]] = None,
    ):
        self._classes: Dict[
            str, Type[LLMProvider]
        ] = {}
        self._instances: Dict[str, LLMProvider] = {}
        self._lazy_map: Dict[str, tuple] = (
            dict(lazy_map) if lazy_map else {}
        )

    def register(
        self,
        name: str,
        cls: Type[LLMProvider],
    ) -> None:
        """Register a provider class under the given name."""
        self._classes[name] = cls

    def register_lazy(
        self,
        name: str,
        module_path: str,
        class_name: str,
    ) -> None:
        """Register a provider for lazy import."""
        self._lazy_map[name] = (module_path, class_name)

    @property
    def available(self) -> List[str]:
        """Return list of registered provider names."""
        names = set(self._classes.keys())
        names.update(self._lazy_map.keys())
        return sorted(names)

    def _resolve_class(
        self, name: str
    ) -> Type[LLMProvider]:
        """Resolve a provider class, importing lazily."""
        if name in self._classes:
            return self._classes[name]
        if name in self._lazy_map:
            module_path, class_name = self._lazy_map[name]
            mod = importlib.import_module(
                module_path, package=__package__
            )
            cls = getattr(mod, class_name)
            self._classes[name] = cls
            return cls
        raise ValueError(
            f"Unknown LLM provider: {name!r}. "
            f"Available: {self.available}"
        )

    def create(self, name: str) -> LLMProvider:
        """Get or create a cached provider instance.

        Parameters
        ----------
        name : str
            The registered provider name.

        Returns
        -------
        LLMProvider
            A (cached) instance of the requested provider.

        Raises
        ------
        ValueError
            If the provider name is not registered.
        """
        if name in self._instances:
            return self._instances[name]
        cls = self._resolve_class(name)
        instance = cls()
        self._instances[name] = instance
        return instance


# Module-level singleton registry (lazy -- no provider
# modules imported until first use)
_registry = ProviderRegistry(lazy_map=_PROVIDER_MAP)


def get_provider(name: str = "") -> LLMProvider:
    """Get a provider instance by name.

    If name is empty, reads LLM_PROVIDER from the environment
    (via akande.config), defaulting to 'openai'.

    Parameters
    ----------
    name : str, optional
        The provider name. If empty, uses config.

    Returns
    -------
    LLMProvider
        An instance of the requested provider.
    """
    if not name:
        from akande.config import LLM_PROVIDER

        name = LLM_PROVIDER or DEFAULT_PROVIDER

    logging.info(
        "LLM provider resolved",
        extra={
            "event": "Provider:Resolved",
            "extra_data": {
                "provider": name,
                "available": _registry.available,
            },
        },
    )
    return _registry.create(name)



# Copyright (C) 2024 Sebastien Rousseau.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import cherrypy
from .cache import SQLiteCache
from .config import OPENAI_DEFAULT_MODEL
from .services import SYSTEM_PROMPT, OpenAIService
from .utils import (
    generate_pdf,
    generate_csv,
    get_output_directory,
    get_output_filename,
)

from concurrent.futures import ThreadPoolExecutor
from functools import partial
import asyncio
import hashlib
import logging
import pyttsx4
import threading
import uuid
import speech_recognition as sr

MAX_THREAD_WORKERS = 4
TTS_RATE = 161
CACHE_DB_NAME = "akande_cache.db"


# Define ANSI escape codes for colors
class Colors:
    RESET = "\033[0m"
    HEADER = "\033[95m"
    RED_BACKGROUND = "\033[48;2;179;0;15m"
    CYAN_BACKGROUND = "\033[48;2;65;175;220m"
    GREEN_BACKGROUND = "\033[48;2;0;103;0m"
    BLUE_BACKGROUND = "\033[48;2;0;78;203m"
    ORANGE_BACKGROUND = "\033[48;2;150;61;0m"


# ANSI escape sequence to clear terminal (replaces subprocess call)
CLEAR_SCREEN = "\033[2J\033[H"


class Akande:
    """
    The Akande voice assistant.

    This class represents the voice assistant capable of understanding
    and responding to user queries. It integrates speech recognition
    and synthesis, leveraging configurable LLM providers for
    generating responses.
    """

    def __init__(self, openai_service: OpenAIService):
        self.server = None
        self.server_thread = None
        self.server_running = False

        # Use a stable cache DB path (date-based dir, fixed name)
        directory_path = get_output_directory()
        cache_path = directory_path / CACHE_DB_NAME

        self.openai_service = openai_service
        self.recognizer = sr.Recognizer()
        self.cache = SQLiteCache(cache_path)
        self.executor = ThreadPoolExecutor(
            max_workers=MAX_THREAD_WORKERS
        )

        # Initialize TTS engine once
        try:
            self._tts_engine = pyttsx4.init()
            self._tts_engine.setProperty("rate", TTS_RATE)
            logging.info(
                "TTS engine initialized",
                extra={"event": "TTS:Initialized"},
            )
        except Exception as e:
            logging.error(
                f"Failed to init TTS engine: "
                f"{type(e).__name__}",
                exc_info=True,
                extra={"event": "TTS:InitFailed"},
            )
            self._tts_engine = None

    def hash_prompt(self, prompt: str) -> str:
        """Hash the prompt for caching."""
        return hashlib.sha256(
            prompt.encode("utf-8")
        ).hexdigest()

    async def speak(self, text: str) -> None:
        """
        Speak the given text using pyttsx4 in an async manner.

        Uses a ThreadPoolExecutor to run blocking TTS operations
        in a separate thread.
        """

        def tts_engine_run(text: str):
            directory_path = get_output_directory()
            filename = get_output_filename(".wav")
            file_path = directory_path / filename
            try:
                if self._tts_engine is not None:
                    self._tts_engine.save_to_file(
                        text, str(file_path)
                    )
                    self._tts_engine.runAndWait()
                    logging.info(
                        "TTS synthesis completed",
                        extra={
                            "event": "Speech:SynthesisCompleted",
                            "extra_data": {
                                "wav_file": str(file_path),
                                "text_length": len(text),
                            },
                        },
                    )
                else:
                    logging.warning(
                        "TTS engine not available",
                        extra={
                            "event": "TTS:Unavailable",
                        },
                    )
            except Exception as e:
                logging.error(
                    f"Error during speech synthesis: "
                    f"{type(e).__name__}",
                    exc_info=True,
                    extra={"event": "Speech:SynthesisFailed"},
                )

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(
            self.executor, partial(tts_engine_run, text)
        )

    async def listen(self) -> str:
        """Listen for user input and return the recognized text."""

        def _listen_sync():
            try:
                with sr.Microphone() as source:
                    logging.info(
                        "Listening for user input",
                        extra={
                            "event": "Speech:RecognitionStarted",
                        },
                    )
                    audio = self.recognizer.listen(source)
                text = self.recognizer.recognize_google(audio)
                logging.info(
                    "Speech recognized",
                    extra={
                        "event": "Speech:RecognitionCompleted",
                        "extra_data": {
                            "success": True,
                            "transcript_length": len(text),
                        },
                    },
                )
                return text
            except sr.UnknownValueError:
                logging.warning(
                    "Speech could not be understood",
                    extra={
                        "event": "Speech:RecognitionCompleted",
                        "extra_data": {
                            "success": False,
                            "error_type": "UnknownValueError",
                        },
                    },
                )
                return ""
            except sr.RequestError as e:
                logging.error(
                    f"Speech recognition service error: "
                    f"{type(e).__name__}",
                    exc_info=True,
                    extra={
                        "event": "Speech:RecognitionCompleted",
                        "extra_data": {
                            "success": False,
                            "error_type": "RequestError",
                        },
                    },
                )
                return ""

        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(
            self.executor, _listen_sync
        )
        if not result:
            await self.speak(
                "I'm sorry, I couldn't understand what you said."
            )
        return result

    async def run_interaction(self) -> None:
        """Main interaction loop of the voice assistant."""
        while True:
            # Clear screen using ANSI escape (no subprocess)
            print(CLEAR_SCREEN, end="", flush=True)
            banner_text = "Àkàndé Voice Assistant"
            banner_width = len(banner_text) + 4
            print(f"{Colors.RESET}{' ' * banner_width}")
            print("  " + banner_text + "  ")
            print(" " * banner_width + Colors.RESET)

            options = [
                ("1. Use voice", Colors.BLUE_BACKGROUND),
                ("2. Ask a question", Colors.GREEN_BACKGROUND),
                ("3. Start server", Colors.ORANGE_BACKGROUND),
                ("4. Stop", Colors.RED_BACKGROUND),
            ]

            for option_text, color in options:
                print(
                    f"{color}{' ' * banner_width}{Colors.RESET}"
                )
                print(
                    f"{color}{option_text:<{banner_width}}"
                    f"{Colors.RESET}"
                )
                print(
                    f"{color}{' ' * banner_width}{Colors.RESET}"
                )

            choice = input(
                "\nPlease select an option: "
            ).strip()

            # Generate correlation ID for this interaction
            correlation_id = str(uuid.uuid4())
            logging.info(
                f"Menu option selected: {choice}",
                extra={
                    "event": "Interaction:MenuSelected",
                    "correlation_id": correlation_id,
                    "extra_data": {"choice": choice},
                },
            )

            if choice == "4":
                print("\nGoodbye!")
                await self.stop_server()
                break
            elif choice == "3":
                await self.run_server()
            elif choice == "2":
                question = input(
                    "Please enter your question: "
                ).strip()
                if question:
                    print("Processing question...")
                    response = await self.generate_response(
                        question,
                        correlation_id=correlation_id,
                    )
                    await self.speak(response)
                    # Offload file generation to background thread
                    loop = asyncio.get_running_loop()
                    loop.run_in_executor(
                        self.executor,
                        generate_pdf,
                        question,
                        response,
                    )
                    loop.run_in_executor(
                        self.executor,
                        generate_csv,
                        question,
                        response,
                    )
                else:
                    print("No question provided.")
            elif choice == "1":
                print("Listening...")
                prompt = (await self.listen()).lower()
                if prompt == "stop":
                    print("\nGoodbye!")
                    await self.stop_server()
                    break
                elif prompt:
                    print("Processing voice command...")
                    response = await self.generate_response(
                        prompt,
                        correlation_id=correlation_id,
                    )
                    await self.speak(response)
                    # Offload file generation to background thread
                    loop = asyncio.get_running_loop()
                    loop.run_in_executor(
                        self.executor,
                        generate_pdf,
                        prompt,
                        response,
                    )
                    loop.run_in_executor(
                        self.executor,
                        generate_csv,
                        prompt,
                        response,
                    )
                else:
                    print("No voice command detected.")
            else:
                print(
                    "Invalid choice. Please select a valid option."
                )

    async def run_server(self) -> None:
        """Run the CherryPy server in a separate thread."""
        if self.server_running:
            logging.info(
                "Server is already running",
                extra={"event": "Server:AlreadyRunning"},
            )
            return

        def start_server():
            from .server.server import AkandeServer

            cherrypy.quickstart(AkandeServer())

        self.server_running = True
        self.server_thread = threading.Thread(
            target=start_server, daemon=True
        )
        self.server_thread.start()
        logging.info(
            "CherryPy server started",
            extra={
                "event": "Server:Started",
                "extra_data": {"port": 8080},
            },
        )

    async def stop_server(self) -> None:
        """Stop the CherryPy server."""
        self.server_running = False
        cherrypy.engine.exit()
        logging.info(
            "CherryPy server stopped",
            extra={"event": "Server:Stopped"},
        )

    async def generate_response(
        self,
        prompt: str,
        correlation_id: str = "",
    ) -> str:
        """
        Generate a response using the LLM provider or cache.

        Args:
            prompt: The prompt for generating the response.
            correlation_id: Optional correlation ID for tracing.

        Returns:
            The generated response.
        """
        prompt_hash = self.hash_prompt(prompt)
        cached_response = self.cache.get(prompt_hash)
        if cached_response:
            logging.info(
                "Using cached response",
                extra={
                    "event": "Response:CacheHit",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "prompt_hash": prompt_hash[:12],
                    },
                },
            )
            return cached_response
        else:
            logging.info(
                "Cache miss, calling LLM provider",
                extra={
                    "event": "Response:CacheMiss",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "prompt_hash": prompt_hash[:12],
                    },
                },
            )
            try:
                response = (
                    await self.openai_service.generate_response(
                        prompt,
                        SYSTEM_PROMPT,
                        OPENAI_DEFAULT_MODEL,
                        {},
                    )
                )
                if not hasattr(response, "choices"):
                    logging.error(
                        "LLM returned unexpected response type",
                        extra={
                            "event": "LLM:UnexpectedResponse",
                            "correlation_id": correlation_id,
                            "extra_data": {
                                "response_type": type(
                                    response
                                ).__name__,
                            },
                        },
                    )
                    return ""
                text_response = (
                    response.choices[0].message.content.strip()
                    if response.choices
                    else ""
                )
                self.cache.set(prompt_hash, text_response)
                return text_response
            except Exception as e:
                logging.error(
                    f"LLM API error: "
                    f"{type(e).__name__}: {e}",
                    exc_info=True,
                    extra={
                        "event": "LLM:Error",
                        "correlation_id": correlation_id,
                    },
                )
                return ""



import cherrypy
import hashlib
import json
import logging
import os
import io
import tempfile
import time
import threading
import uuid
import speech_recognition as sr
from pathlib import Path
from pydub import AudioSegment
from pydub.exceptions import CouldntDecodeError
from akande.cache import SQLiteCache
from akande.config import (
    LLM_PROVIDER,
    OPENAI_API_KEY,
    OPENAI_DEFAULT_MODEL,
)
from akande.providers import get_provider
from akande.services import SYSTEM_PROMPT, OpenAIImpl
from akande.utils import (
    validate_api_key,
    get_output_directory,
)

ALLOWED_STATIC_FILES = {"sine-wave-generator.js"}
MAX_QUESTION_LENGTH = 5000
MAX_AUDIO_SIZE = 10 * 1024 * 1024  # 10 MB
RATE_LIMIT_WINDOW = 60  # seconds
RATE_LIMIT_MAX_REQUESTS = 20  # per window per IP
CACHE_DB_NAME = "akande_cache.db"

# Magic bytes for audio format detection
AUDIO_SIGNATURES = {
    b"\x1a\x45\xdf\xa3": "webm",
    b"ID3": "mp3",
    b"\xff\xfb": "mp3",
    b"\xff\xf3": "mp3",
    b"\xff\xf2": "mp3",
    b"OggS": "ogg",
    b"fLaC": "flac",
}

# Module-level shared recognizer instance
_recognizer = sr.Recognizer()


def _detect_audio_format(data: bytes) -> str:
    """Detect audio format from magic bytes."""
    for signature, fmt in AUDIO_SIGNATURES.items():
        if data[: len(signature)] == signature:
            return fmt
    # Check for mp4 (ftyp at offset 4)
    if len(data) > 8 and data[4:8] == b"ftyp":
        return "mp4"
    return ""


class RateLimiter:
    """Thread-safe in-memory per-IP rate limiter."""

    def __init__(self, window: int, max_requests: int):
        self.window = window
        self.max_requests = max_requests
        self._requests: dict = {}
        self._lock = threading.Lock()

    def is_allowed(self, ip: str) -> bool:
        now = time.time()
        cutoff = now - self.window
        with self._lock:
            timestamps = self._requests.get(ip, [])
            timestamps = [t for t in timestamps if t > cutoff]
            if len(timestamps) >= self.max_requests:
                self._requests[ip] = timestamps
                return False
            timestamps.append(now)
            self._requests[ip] = timestamps
        return True

    def cleanup(self):
        """Remove stale IPs with no recent requests."""
        cutoff = time.time() - self.window
        with self._lock:
            stale = [
                ip
                for ip, ts in self._requests.items()
                if not any(t > cutoff for t in ts)
            ]
            for ip in stale:
                del self._requests[ip]


def _hash_ip(ip: str) -> str:
    """Hash an IP address for logging (PII protection)."""
    return hashlib.sha256(ip.encode()).hexdigest()[:12]


class SecurityHeadersTool(cherrypy.Tool):
    """CherryPy tool to add security headers to all responses."""

    def __init__(self):
        super().__init__(
            "before_finalize", self._set_headers
        )

    def _set_headers(self):
        h = cherrypy.response.headers
        h["X-Content-Type-Options"] = "nosniff"
        h["X-Frame-Options"] = "DENY"
        h["Referrer-Policy"] = (
            "strict-origin-when-cross-origin"
        )
        h["Permissions-Policy"] = "microphone=(self)"
        h["X-XSS-Protection"] = "1; mode=block"
        h["Content-Security-Policy"] = (
            "default-src 'self'; "
            "script-src 'self' 'unsafe-inline' "
            "*.google-analytics.com "
            "https://cdn.jsdelivr.net "
            "www.googletagmanager.com "
            "x.clarity.ms "
            "https://www.googletagmanager.com "
            "https://www.google.com "
            "https://www.gstatic.com; "
            "frame-src 'self' https://www.google.com; "
            "connect-src 'self' www.googletagmanager.com "
            "https://region1.google-analytics.com; "
            "img-src 'self' data: https: blob: kura.pro "
            "www.googletagmanager.com; "
            "style-src 'self' 'unsafe-inline' "
            "https://cdn.jsdelivr.net "
            "https://fonts.googleapis.com "
            "https://cdnjs.cloudflare.com "
            "https://use.fontawesome.com; "
            "font-src 'self' https://use.fontawesome.com/; "
            "media-src 'self';"
        )


# Register security headers tool
cherrypy.tools.security_headers = SecurityHeadersTool()


class AkandeServer:
    _rate_limiter = RateLimiter(
        RATE_LIMIT_WINDOW, RATE_LIMIT_MAX_REQUESTS
    )

    _cp_config = {
        "tools.security_headers.on": True,
    }

    def __init__(self):
        provider_name = LLM_PROVIDER or "openai"
        if provider_name == "openai":
            if not validate_api_key(OPENAI_API_KEY):
                raise RuntimeError(
                    "Invalid or missing OPENAI_API_KEY. "
                    "Server cannot start without a "
                    "valid API key."
                )
            self.openai_service = OpenAIImpl()
        else:
            self.openai_service = get_provider(
                provider_name
            )
        self.logger = logging.getLogger(__name__)
        self.public_dir = (
            Path(__file__).resolve().parent.parent.parent
            / "public"
        )
        # Server-side cache
        directory_path = get_output_directory()
        cache_path = directory_path / CACHE_DB_NAME
        self.cache = SQLiteCache(cache_path)

        self.logger.info(
            "Server initialized",
            extra={
                "event": "Server:Initialized",
                "extra_data": {
                    "public_dir": str(self.public_dir),
                },
            },
        )

    def _get_correlation_id(self) -> str:
        """Get or generate a correlation ID for the request."""
        return cherrypy.request.headers.get(
            "X-Request-Id", str(uuid.uuid4())
        )

    def _check_rate_limit(self):
        ip = cherrypy.request.remote.ip
        if not self._rate_limiter.is_allowed(ip):
            ip_hash = _hash_ip(ip)
            self.logger.warning(
                "Rate limit exceeded",
                extra={
                    "event": "Server:RateLimited",
                    "extra_data": {
                        "ip_hash": ip_hash,
                        "window": RATE_LIMIT_WINDOW,
                        "max_requests": RATE_LIMIT_MAX_REQUESTS,
                    },
                },
            )
            cherrypy.response.status = 429
            raise cherrypy.HTTPError(
                429, "Rate limit exceeded. Try again later."
            )

    @cherrypy.expose
    def health(self):
        """Health check endpoint."""
        cherrypy.response.headers["Content-Type"] = (
            "application/json"
        )
        return json.dumps(
            {"status": "ok", "service": "akande"}
        ).encode("utf-8")

    @cherrypy.expose
    def index(self):
        index_path = self.public_dir / "index.html"
        if not index_path.is_file():
            raise cherrypy.HTTPError(404, "Index page not found")
        return index_path.read_text(encoding="utf-8")

    @cherrypy.expose
    def static(self, path):
        if path not in ALLOWED_STATIC_FILES:
            ip_hash = _hash_ip(
                cherrypy.request.remote.ip
            )
            self.logger.warning(
                "Forbidden static file access attempt",
                extra={
                    "event": "Server:ForbiddenAccess",
                    "extra_data": {
                        "ip_hash": ip_hash,
                        "path": path[:100],
                    },
                },
            )
            raise cherrypy.HTTPError(403, "Forbidden")
        file_path = self.public_dir / path
        resolved = file_path.resolve()
        if not str(resolved).startswith(
            str(self.public_dir.resolve())
        ):
            raise cherrypy.HTTPError(403, "Forbidden")
        if not resolved.is_file():
            raise cherrypy.HTTPError(404, "File not found")
        return resolved.read_text(encoding="utf-8")

    @cherrypy.expose
    @cherrypy.tools.allow(methods=["POST"])
    def process_question(self):
        self._check_rate_limit()
        correlation_id = self._get_correlation_id()
        start_time = time.time()
        try:
            request_data = json.loads(
                cherrypy.request.body.read()
            )
            question = request_data.get("question", "")

            if (
                not isinstance(question, str)
                or not question.strip()
            ):
                cherrypy.response.status = 400
                return json.dumps(
                    {
                        "error": (
                            "Question must be a "
                            "non-empty string"
                        )
                    }
                )

            question = question.strip()[:MAX_QUESTION_LENGTH]

            self.logger.info(
                "Text question received",
                extra={
                    "event": "Server:RequestReceived",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "method": "POST",
                        "path": "/process_question",
                        "question_length": len(question),
                    },
                },
            )

            # Check cache first
            prompt_hash = hashlib.sha256(
                question.encode("utf-8")
            ).hexdigest()
            cached = self.cache.get(prompt_hash)
            if cached:
                latency = (time.time() - start_time) * 1000
                self.logger.info(
                    "Served from cache",
                    extra={
                        "event": "Server:RequestCompleted",
                        "correlation_id": correlation_id,
                        "extra_data": {
                            "status": 200,
                            "cache_hit": True,
                            "latency_ms": round(latency, 2),
                        },
                    },
                )
                return json.dumps({"response": cached})

            response_object = (
                self.openai_service.generate_response_sync(
                    question,
                    SYSTEM_PROMPT,
                    OPENAI_DEFAULT_MODEL,
                    None,
                )
            )
            message_content = (
                response_object.choices[0].message.content
            )
            # Store in cache
            self.cache.set(prompt_hash, message_content)

            latency = (time.time() - start_time) * 1000
            self.logger.info(
                "Text question processed",
                extra={
                    "event": "Server:RequestCompleted",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "status": 200,
                        "cache_hit": False,
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            return json.dumps({"response": message_content})

        except json.JSONDecodeError:
            self.logger.warning(
                "Invalid JSON in request",
                extra={
                    "event": "Server:BadRequest",
                    "correlation_id": correlation_id,
                },
            )
            cherrypy.response.status = 400
            return json.dumps({"error": "Invalid JSON"})
        except Exception as e:
            latency = (time.time() - start_time) * 1000
            self.logger.error(
                f"Failed to process question: "
                f"{type(e).__name__}",
                exc_info=True,
                extra={
                    "event": "Server:RequestFailed",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            cherrypy.response.status = 500
            return json.dumps(
                {"response": "An error occurred"}
            )

    @cherrypy.expose
    @cherrypy.tools.allow(methods=["POST"])
    def process_audio_question(self):
        self._check_rate_limit()
        correlation_id = self._get_correlation_id()
        start_time = time.time()
        try:
            audio_data = cherrypy.request.body.read()

            if len(audio_data) > MAX_AUDIO_SIZE:
                cherrypy.response.status = 400
                return json.dumps(
                    {"error": "Audio file too large"}
                )

            if len(audio_data) == 0:
                cherrypy.response.status = 400
                return json.dumps(
                    {"error": "No audio data received"}
                )

            # Use Content-Type hint from browser if available
            content_type = cherrypy.request.headers.get(
                "Content-Type", ""
            )

            self.logger.info(
                "Audio question received",
                extra={
                    "event": "Server:RequestReceived",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "method": "POST",
                        "path": "/process_audio_question",
                        "audio_size": len(audio_data),
                        "content_type": content_type,
                    },
                },
            )

            wav_file_path = self.convert_to_wav(
                audio_data, content_type, correlation_id
            )
            try:
                processed_result = self.process_audio(
                    wav_file_path, correlation_id
                )

                if not processed_result.get("success"):
                    return json.dumps(
                        {
                            "error": processed_result.get(
                                "error",
                                "Audio processing failed",
                            )
                        }
                    )

                question = processed_result.get("text", "")
                if not question:
                    return json.dumps(
                        {"error": "No speech detected"}
                    )

                # Check cache first
                prompt_hash = hashlib.sha256(
                    question.encode("utf-8")
                ).hexdigest()
                cached = self.cache.get(prompt_hash)
                if cached:
                    latency = (
                        time.time() - start_time
                    ) * 1000
                    self.logger.info(
                        "Audio question served from cache",
                        extra={
                            "event": (
                                "Server:RequestCompleted"
                            ),
                            "correlation_id": correlation_id,
                            "extra_data": {
                                "status": 200,
                                "cache_hit": True,
                                "latency_ms": round(
                                    latency, 2
                                ),
                            },
                        },
                    )
                    return json.dumps(
                        {"response": cached}
                    )

                response_object = (
                    self.openai_service.generate_response_sync(
                        question,
                        SYSTEM_PROMPT,
                        OPENAI_DEFAULT_MODEL,
                        None,
                    )
                )
                message_content = (
                    response_object.choices[0].message.content
                )
                self.cache.set(prompt_hash, message_content)

                latency = (time.time() - start_time) * 1000
                self.logger.info(
                    "Audio question processed",
                    extra={
                        "event": "Server:RequestCompleted",
                        "correlation_id": correlation_id,
                        "extra_data": {
                            "status": 200,
                            "cache_hit": False,
                            "latency_ms": round(latency, 2),
                        },
                    },
                )
                return json.dumps(
                    {"response": message_content}
                )
            finally:
                if os.path.exists(wav_file_path):
                    os.remove(wav_file_path)

        except Exception as e:
            latency = (time.time() - start_time) * 1000
            self.logger.error(
                f"Failed to process audio: "
                f"{type(e).__name__}",
                exc_info=True,
                extra={
                    "event": "Server:RequestFailed",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            cherrypy.response.status = 500
            return json.dumps(
                {"error": "Failed to process audio"}
            ).encode("utf-8")

    @staticmethod
    def convert_to_wav(
        audio_data,
        content_type="",
        correlation_id="",
    ):
        convert_start = time.time()
        logger = logging.getLogger(__name__)
        try:
            audio_segment = None

            # Try Content-Type hint first
            ct_format_map = {
                "audio/webm": "webm",
                "audio/mp4": "mp4",
                "audio/mpeg": "mp3",
                "audio/ogg": "ogg",
                "audio/flac": "flac",
            }
            ct_fmt = ct_format_map.get(content_type, "")
            if ct_fmt:
                try:
                    audio_segment = AudioSegment.from_file(
                        io.BytesIO(audio_data), format=ct_fmt
                    )
                except CouldntDecodeError:
                    pass

            # Try magic-byte detection if Content-Type didn't work
            if audio_segment is None:
                detected = _detect_audio_format(audio_data)
                if detected:
                    try:
                        audio_segment = AudioSegment.from_file(
                            io.BytesIO(audio_data),
                            format=detected,
                        )
                    except CouldntDecodeError:
                        pass

            # Fall back to brute-force
            if audio_segment is None:
                for fmt in [
                    "webm",
                    "mp3",
                    "mp4",
                    "ogg",
                    "flac",
                ]:
                    try:
                        audio_segment = AudioSegment.from_file(
                            io.BytesIO(audio_data), format=fmt
                        )
                        break
                    except CouldntDecodeError:
                        pass

            if audio_segment is None:
                raise ValueError("Unsupported audio format")

            audio_segment = audio_segment.set_channels(
                1
            ).set_frame_rate(16000)

            tmp = tempfile.NamedTemporaryFile(
                suffix=".wav", delete=False
            )
            tmp.close()
            audio_segment.export(tmp.name, format="wav")

            latency = (time.time() - convert_start) * 1000
            logger.info(
                "Audio converted to WAV",
                extra={
                    "event": "Server:AudioConverted",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "input_size": len(audio_data),
                        "content_type": content_type,
                        "latency_ms": round(latency, 2),
                    },
                },
            )
            return tmp.name

        except Exception as e:
            logger.error(
                f"Audio conversion failed: "
                f"{type(e).__name__}",
                exc_info=True,
                extra={
                    "event": "Server:AudioConversionFailed",
                    "correlation_id": correlation_id,
                },
            )
            raise RuntimeError(f"Error converting audio: {e}")

    @staticmethod
    def process_audio(file_path, correlation_id=""):
        logger = logging.getLogger(__name__)
        try:
            with sr.AudioFile(file_path) as source:
                audio_data = _recognizer.record(source)

            text = _recognizer.recognize_google(audio_data)
            logger.info(
                "Audio speech recognized",
                extra={
                    "event": "Speech:RecognitionCompleted",
                    "correlation_id": correlation_id,
                    "extra_data": {
                        "success": True,
                        "transcript_length": len(text),
                    },
                },
            )
            return {"text": text, "success": True}

        except sr.UnknownValueError:
            logger.warning(
                "Audio speech not understood",
                extra={
                    "event": "Speech:RecognitionCompleted",
                    "correlation_id": correlation_id,
                    "extra_data": {"success": False},
                },
            )
            return {
                "error": "Audio could not be understood",
                "success": False,
            }
        except sr.RequestError:
            logger.error(
                "Speech recognition service error",
                exc_info=True,
                extra={
                    "event": "Speech:RecognitionFailed",
                    "correlation_id": correlation_id,
                },
            )
            return {
                "error": "Speech recognition service error",
                "success": False,
            }


def main():
    # Only configure logging if no handlers exist yet
    if not logging.getLogger().handlers:
        logging.basicConfig(level=logging.INFO)

    cherrypy.config.update(
        {
            "server.socket_host": "127.0.0.1",
            "server.socket_port": 8080,
            "server.thread_pool": 30,
            "server.max_request_body_size": MAX_AUDIO_SIZE,
            "request.show_tracebacks": False,
            "request.show_mismatched_params": False,
        }
    )
    cherrypy.quickstart(AkandeServer())


if __name__ == "__main__":
    main()
